""" reservoir.py: This module contains classes to simulate the operation of a reservoir.
"""

import math
import os

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from pownet.data_utils import remove_29feb
from pownet.folder_utils import get_model_dir, get_reservoir_file
from .solve_release import (
    solve_release_from_target_storage,
    solve_release_from_dispatch,
)
from .reservoir_functions import (
    calc_min_flow,
    calc_target_level,
    calc_target_storage,
    calc_level_from_storage,
    calc_daily_hydropower,
    calc_release_impact,
    calc_max_release,
    calc_min_release,
    convert_to_hourly_hydropower,
)


class Reservoir:
    """This class simulates a reservoir."""

    def __init__(
        self,
        model_name: str,
        reservoir_name: str,
        upstream_flow: pd.Series = None,
        sim_horizon: int = 365,
    ) -> None:

        # Number of days in a year
        self.sim_horizon = sim_horizon

        self.model_name: str = model_name
        self.name: str = reservoir_name

        # Upstream flow
        self.upstream_flow: pd.Series = None
        if upstream_flow is not None:
            self.set_upstream_flow(upstream_flow)
        else:
            self.upstream_flow: pd.Series = pd.Series(
                0, index=range(1, self.sim_horizon + 1)
            )

        ############################
        # Loaded parameters from CSV
        ############################

        self.min_day: int = None
        self.max_day: int = None
        self.min_level: float = None  # in meters
        self.max_level: float = None  # in meters

        # Max head is the maximum distance between the water level and turbine
        self.max_head: float = None  # in meters
        self.max_storage: float = None  # in m3
        self.max_release: float = None  # in m3/day
        # Max generation is power in MW
        self.max_generation: float = None
        # Index of the reservoir's cascade level. Zero is the furthest upstream.
        self.cascade_level: int = None
        # Turbine efficiency
        self.turbine_factor: float = None

        self.inflow: pd.Series = None  # in m3/day
        self.mean_annual_flow: pd.Series = None  # in m3/day

        ############################
        # Simulated values
        ############################

        self.level: pd.Series = None  # in meters
        # The mid_level is the average of the level of the current day and the next day
        self.mid_level: pd.Series = None  # in meters
        self.storage: pd.Series = None  # in m3
        self.release: pd.Series = None  # in m3/day
        self.spill: pd.Series = None  # in m3/day
        self.min_flow: float = None  # in m3/day
        self.daily_hydropower: pd.Series = None  # in MW-day

        # These policies are calculated based on the min_day and max_day
        self.target_level: pd.Series = None  # in meters
        self.target_storage: pd.Series = None  # in m3

        ############################
        # Reoperated values
        ############################

        self.reop_upstream_flow: pd.Series = pd.Series(
            np.nan, index=range(1, self.sim_horizon + 1)
        )
        self.reop_storage: pd.Series = pd.Series(
            np.nan, index=range(1, self.sim_horizon + 1)
        )
        self.reop_release: pd.Series = pd.Series(
            np.nan, index=range(1, self.sim_horizon + 1)
        )
        self.reop_spill: pd.Series = pd.Series(
            np.nan, index=range(1, self.sim_horizon + 1)
        )
        self.reop_level: pd.Series = pd.Series(
            np.nan, index=range(1, self.sim_horizon + 1)
        )
        self.reop_daily_hydropower: pd.Series = pd.Series(
            np.nan, index=range(1, self.sim_horizon + 1)
        )

    def set_upstream_flow(self, upstream_flow: pd.Series) -> None:
        """Set the upstream flow of the reservoir."""
        # Check the index of upstream_flow is between 1 and sim_horizon
        if (upstream_flow.index[0] != 1) or (
            upstream_flow.index[-1] != self.sim_horizon
        ):
            raise ValueError("Index of upstream flow should be in [1, 365].")
        # Length of upstream flow should be equal to the simulation horizon
        if len(upstream_flow) != self.sim_horizon:
            raise ValueError(
                "Length of upstream flow should be equal to the simulation horizon."
            )
        self.upstream_flow = upstream_flow

    def load_flows(self) -> None:
        """Load inflow and mean annual flow from CSV files.
        The files are located in model_library/model_name/reservoir/.
        """
        # Inflow (m3/day)
        inflow_data = pd.read_csv(
            get_reservoir_file(model_name=self.model_name, filename="inflow.csv")
        )
        self.inflow = remove_29feb(inflow_data)
        self.inflow = self.inflow[self.name]

        # Mean annual flow (m3/day)
        mean_annual_flow_data = pd.read_csv(
            get_reservoir_file(
                model_name=self.model_name, filename="mean_annual_flow.csv"
            )
        )
        self.mean_annual_flow = remove_29feb(mean_annual_flow_data)
        self.mean_annual_flow = self.mean_annual_flow[self.name]

    def load_reservoir_params(self) -> None:
        """Load reservoir parameters from reservoir.csv."""
        reservoir_params = (
            pd.read_csv(
                os.path.join(
                    get_model_dir(), self.model_name, "reservoir", "reservoir.csv"
                ),
                header=0,
            )
            .set_index("name")
            .loc[self.name]
        )
        self.min_day = int(reservoir_params["min_day"])
        self.max_day = int(reservoir_params["max_day"])
        self.min_level = reservoir_params["min_level"]
        self.max_level = reservoir_params["max_level"]
        self.max_head = reservoir_params["max_head"]
        self.max_storage = reservoir_params["max_storage"]
        self.max_release = reservoir_params["max_release"]
        self.max_generation = reservoir_params["max_generation"]
        self.cascade_level = reservoir_params["cascade_level"]
        self.turbine_factor = reservoir_params["turbine_factor"]

    def load_from_csv(
        self,
    ) -> None:
        """Load reservoir parameters from /model_name/reservoir/reservoir.csv.

        Args:
            model_name (str): The name of the model
            reservoir_name (str): The name of the reservoir
            upstream_flow (pd.Series): The inflow from upstream reservoirs

        Returns:
            None
        """
        self.load_flows()
        self.load_reservoir_params()

    def simulate(self):
        """Simulate the operation of the reservoir."""
        # Minimum flow
        self.min_flow = calc_min_flow(
            inflow=self.inflow,
            mean_annual_flow=self.mean_annual_flow,
            max_release=self.max_release,
        )
        # Calculate the target storage and level
        self.target_level = calc_target_level(
            min_day=self.min_day,
            max_day=self.max_day,
            min_level=self.min_level,
            max_level=self.max_level,
        )
        self.target_storage = calc_target_storage(
            target_level=self.target_level,
            min_level=self.min_level,
            max_level=self.max_level,
            max_storage=self.max_storage,
        )

        # Get reservoir timeseries values
        self.release, self.spill, self.storage, _ = solve_release_from_target_storage(
            reservoir_name=self.name,
            start_day=1,
            end_day=self.sim_horizon,
            max_release=self.max_release,
            max_storage=self.max_storage,
            initial_storage=self.target_storage[
                1
            ],  # Assume the initial storage is the target storage of the first day
            target_storage=self.target_storage,
            min_flow=self.min_flow,
            total_inflow=self.inflow + self.upstream_flow,
        )

        self.level = calc_level_from_storage(
            storage=self.storage,
            min_level=self.min_level,
            max_level=self.max_level,
            max_storage=self.max_storage,
        )

        self.mid_level = (self.level + self.level.shift(1)) / 2
        # Assume the mid_level of the first day is the target level of the first day
        self.mid_level[1] = self.target_level[1]

        self.daily_hydropower = calc_daily_hydropower(
            release=self.release,
            mid_level=self.mid_level,
            max_generation=self.max_generation,
            turbine_factor=self.turbine_factor,
            max_head=self.max_head,
            max_level=self.max_level,
        )

    def reoperate(
        self, day: int, daily_dispatch: float, upstream_flow: float
    ) -> pd.DataFrame:
        """Reoperate the reservoir based on the daily dispatch of the power system model.
        There are seven cases which are outlined in the code.
        Note t-1 is denoted as t0; t is denoted as t; t+1 is denoted as t1.

        Args:
            day (int): The current day for which the reoperation is being calculated.
            daily_dispatch (float): The daily dispatch value from the power system model.
            upstream_flow (float): The flow of water from upstream.

        Returns:
            pd.DataFrame: A DataFrame containing the reoperation results for the reservoir.
        """
        # Upstream flow may change every reoperation iteration because
        # reservoirs located upstream may reoperate.
        self.reop_upstream_flow.loc[day] = upstream_flow

        # Since release is fixed on the first day, we do not reoperate and
        # return the original dispatch.
        if day == 1:
            self.reop_release.loc[day] = self.release.loc[day]
            self.reop_spill.loc[day] = self.spill.loc[day]
            self.reop_storage.loc[day] = self.storage.loc[day]
            self.reop_level.loc[day] = calc_level_from_storage(
                storage=self.reop_storage.loc[day],
                min_level=self.min_level,
                max_level=self.max_level,
                max_storage=self.max_storage,
            )
            self.reop_daily_hydropower.loc[day] = self.daily_hydropower.loc[day]
            return daily_dispatch

        # Values for calculations are from previous reoperation attemps
        total_inflow_t = self.inflow.loc[day] + upstream_flow
        release_t0 = self.reop_release.loc[day - 1]
        storage_t0 = self.reop_storage.loc[day - 1]
        level_t0 = calc_level_from_storage(
            storage=storage_t0,
            min_level=self.min_level,
            max_level=self.max_level,
            max_storage=self.max_storage,
        )
        min_flow_t = self.min_flow.loc[day]

        #########################################################
        ##### Max release and its corresponding values
        #########################################################
        # Change in release is limited to 15% of the maximum release
        hydropeak_factor = 0.15
        max_release_t = calc_max_release(
            total_inflow_t=total_inflow_t,
            release_t0=release_t0,
            storage_t0=storage_t0,
            min_flow_t=min_flow_t,
            max_release=self.max_release,
            hydropeak_factor=hydropeak_factor,
        )
        # Reservoir state due to max_release_t
        (
            spill_from_max_release_t,
            storage_from_max_release_t,
            level_from_max_release_t,
            max_daily_hydropower_t,
        ) = calc_release_impact(
            release_t=max_release_t,
            storage_t0=storage_t0,
            total_inflow_t=total_inflow_t,
            min_level=self.min_level,
            max_level=self.max_level,
            max_storage=self.max_storage,
            level_t0=level_t0,
            max_generation=self.max_generation,
            turbine_factor=self.turbine_factor,
            max_head=self.max_head,
        )

        #########################################################
        ##### Min release and its corresponding values
        #########################################################
        min_release_t = calc_min_release(
            total_inflow_t=total_inflow_t,
            release_t0=release_t0,
            storage_t0=storage_t0,
            min_flow_t=min_flow_t,
            max_release=self.max_release,
            hydropeak_factor=hydropeak_factor,
        )
        # Reservoir state due to min_release_t
        (
            spill_from_min_release_t,
            storage_from_min_release_t,
            level_from_min_release_t,
            min_daily_hydropower_t,
        ) = calc_release_impact(
            release_t=min_release_t,
            storage_t0=storage_t0,
            total_inflow_t=total_inflow_t,
            min_level=self.min_level,
            max_level=self.max_level,
            max_storage=self.max_storage,
            level_t0=level_t0,
            max_generation=self.max_generation,
            turbine_factor=self.turbine_factor,
            max_head=self.max_head,
        )

        ##############################
        ##### Reoperation cases
        ##############################

        # Define the previous guesses of daily_hydropower
        daily_hydropower_rule_curve = self.daily_hydropower.loc[day]
        previous_reop_daily_hydropower = self.reop_daily_hydropower.loc[day]

        # Two values are equal when they are within 0.1% of daily max capacity (MW-day)
        tolerance = 0.001 * self.max_generation * 24

        # Case 1: If dispatch is equal to the previous guess from reoperation,
        # then terminate.
        if math.isclose(
            daily_dispatch, previous_reop_daily_hydropower, abs_tol=tolerance
        ):
            return daily_dispatch

        # Case 2: If dispatch is equal to min daily_hydropower,
        # set release to min release
        elif math.isclose(daily_dispatch, min_daily_hydropower_t, abs_tol=tolerance):
            self.reop_release.loc[day] = min_release_t
            self.reop_spill.loc[day] = spill_from_min_release_t
            self.reop_storage.loc[day] = storage_from_min_release_t
            self.reop_level.loc[day] = level_from_min_release_t
            self.reop_daily_hydropower.loc[day] = min_daily_hydropower_t
            return daily_dispatch

        # Case 3: If dispatch is equal to max daily_hydropower,
        # set release to max release.
        elif math.isclose(daily_dispatch, max_daily_hydropower_t, abs_tol=tolerance):
            self.reop_release.loc[day] = max_release_t
            self.reop_spill.loc[day] = spill_from_max_release_t
            self.reop_storage.loc[day] = storage_from_max_release_t
            self.reop_level.loc[day] = level_from_max_release_t
            self.reop_daily_hydropower.loc[day] = max_daily_hydropower_t
            return daily_dispatch

        # Case 4: Release cannot be less than the minimum flow.
        # Likewise, set release to min release.
        elif daily_dispatch < min_daily_hydropower_t:
            self.reop_release.loc[day] = min_release_t
            self.reop_spill.loc[day] = spill_from_min_release_t
            self.reop_storage.loc[day] = storage_from_min_release_t
            self.reop_level.loc[day] = level_from_min_release_t
            self.reop_daily_hydropower.loc[day] = min_daily_hydropower_t
            # Return the original dispatch to get convergence
            # for the reoperation
            return min_daily_hydropower_t

        # Case 5: Dismatch cannot be greater than max daily_hydropower,
        # therefore set release to its maximum value.
        elif daily_dispatch > max_daily_hydropower_t:
            self.reop_release.loc[day] = max_release_t
            self.reop_spill.loc[day] = spill_from_max_release_t
            self.reop_storage.loc[day] = storage_from_max_release_t
            self.reop_level.loc[day] = level_from_max_release_t
            self.reop_daily_hydropower.loc[day] = max_daily_hydropower_t
            return max_daily_hydropower_t

        # Case 6: If the reoperation is the first iteration, dispatch can be
        # limited by the rule curve. Therefore, the new estimated release
        # should be the maximum release to bypass the limit.
        elif math.isclose(
            daily_dispatch,
            daily_hydropower_rule_curve,
            abs_tol=tolerance,
        ):
            self.reop_release.loc[day] = max_release_t
            self.reop_spill.loc[day] = spill_from_max_release_t
            self.reop_storage.loc[day] = storage_from_max_release_t
            self.reop_level.loc[day] = level_from_max_release_t
            self.reop_daily_hydropower.loc[day] = max_daily_hydropower_t
            return max_daily_hydropower_t

        # Case 7: If dispatch is between min_daily_hydropower and max_daily_hydropower,
        # then solve for release from dispatch using an optimization algorithm.
        elif min_daily_hydropower_t < daily_dispatch < max_daily_hydropower_t:
            (
                reop_release_t,
                reop_spill_t,
                reop_storage_t,
                reop_level_t,
                reop_hourly_hydropower_t,
                reop_daily_hydropower_t,
                reop_mismatch_t,
            ) = solve_release_from_dispatch(
                reservoir_name=self.name,
                dispatch=daily_dispatch,
                turbine_factor=self.turbine_factor,
                max_head=self.max_head,
                max_level=self.max_level,
                min_level=self.min_level,
                level_t0=level_t0,
                storage_max=self.max_storage,
                storage_t0=storage_t0,
                inflow=total_inflow_t,
                min_release=min_release_t,
                max_release=max_release_t,
                max_generation=self.max_generation,
            )

            ##############################
            # Ensure that equations in the optimization problem
            # are correct by comparing the results with values
            # from manual calculations.
            ##############################
            (
                _,
                temp_storage,
                temp_level,
                temp_daily_hydropower,
            ) = calc_release_impact(
                release_t=reop_release_t,
                storage_t0=storage_t0,
                total_inflow_t=total_inflow_t,
                min_level=self.min_level,
                max_level=self.max_level,
                max_storage=self.max_storage,
                level_t0=level_t0,
                max_generation=self.max_generation,
                turbine_factor=self.turbine_factor,
                max_head=self.max_head,
            )

            comparisons = [
                (temp_storage, reop_storage_t),
                (temp_level, reop_level_t),
                (temp_daily_hydropower, reop_daily_hydropower_t),
            ]
            if not all(math.isclose(a, b, rel_tol=tolerance) for a, b in comparisons):
                raise ValueError("Optimization did not produce the correct values.")

            # Update the reoperation values
            self.reop_release.loc[day] = reop_release_t
            self.reop_spill.loc[day] = reop_spill_t
            self.reop_storage.loc[day] = reop_storage_t
            self.reop_level.loc[day] = reop_level_t
            self.reop_daily_hydropower.loc[day] = reop_daily_hydropower_t
            return reop_daily_hydropower_t

        # Catch other cases.
        else:
            raise ValueError(
                f"Unknown case: {daily_dispatch} vs. {daily_hydropower_rule_curve}"
            )

    def plot_state(self, output_folder: str = None) -> None:
        fig, ax = plt.subplots(figsize=(13, 7), layout="constrained", dpi=350)
        ax.plot(self.inflow + self.upstream_flow, label="Inflow (m3/day)")
        ax.plot(self.release, label="Release (m3/day)")
        ax.plot(self.spill, label="Spill (m3/day)", linestyle="dotted", linewidth=2)
        ax.plot(
            self.release + self.spill,
            label="Outflow (m3/day)",
            linestyle="dotted",
        )
        ax.set_xlabel("Day")
        ax.set_ylabel("Flow rate (m3/day)")
        ax.set_title(self.name)

        ax2 = ax.twinx()
        ax2.plot(self.storage, label="Storage (m3)", color="k", linewidth=1)
        ax2.plot(
            self.target_storage,
            label="Target Storage (m3)",
            linestyle="--",
            color="k",
            linewidth=5,
            alpha=0.5,
        )
        ax2.set_ylabel("Storage (m3)")
        fig.legend(loc="outside right upper")

        # Save figure
        if output_folder is not None:
            fig.savefig(os.path.join(output_folder, f"{self.name}.png"))

        plt.show()

    def get_hourly_hydropower(self) -> pd.Series:
        return convert_to_hourly_hydropower(self.daily_hydropower)

    def get_daily_hydropower(self) -> pd.Series:
        """Return the daily hydropower values."""
        return self.daily_hydropower

    def get_reop_daily_hydropower(self) -> pd.Series:
        """Return the reoperated daily hydropower values."""
        return self.reop_daily_hydropower



"""data_utils.py: functions for processing user inputs"""

import re
import datetime
import os

import numpy as np
import geopandas as gpd
import pandas as pd
from shapely.geometry import LineString, Point

from .folder_utils import get_database_dir


def get_dates(year, num_days=365):
    """Return a dataframe of dates for the given year. The dataframe will have
    365 rows, one for each day of the year. The columns are 'date' and 'hour'.
    Exclude 29th February.
    """
    # Create dates to concatenate with the new dataframes
    dates = pd.DataFrame(
        {"date": pd.date_range(start=str(year), periods=num_days + 1, freq="D")}
    )
    # Remove 29th Feb because we do not deal with them
    dates = dates.loc[dates.date.dt.strftime("%m-%d") != "02-29"]
    # Remove 1st Jan of the next year in case it is included when it is not a leap year
    dates = dates.loc[dates.date.dt.strftime("%Y-%m-%d") != f"{year+1}-01-01"]

    # In case we need three columns: date, hour, and day
    dates = dates.loc[dates.index.repeat(24)]
    dates["hour"] = np.tile(range(1, 25), num_days)
    dates = dates.reset_index(drop=True)
    return dates


def get_datetime_index(year: int) -> pd.DatetimeIndex:
    """Return a datetime index for the given year. The index will have 8760 entries, one for each hour of the year. Exclude 29th February."""
    dates = pd.date_range(start=f"{year}-01-01", end=f"{year+1}-01-01", freq="H")
    # Remove 29th February
    dates = dates[~((dates.month == 2) & (dates.day == 29))]
    return dates[dates.year == year]


def remove_29feb(timeseries: pd.Series) -> pd.Series:
    """Clean and reindex flow data by removing February 29th and resetting the index.

    Args:
        timeseries (pd.Series): Timeseries to be cleaned and reindexed.

    Returns:
        pd.Series: The cleaned and reindexed flow data.
    """

    # Remove 29th February from the dataset
    cleaned_data = timeseries[~((timeseries.month == 2) & (timeseries.day == 29))]
    # Index starts from 1
    cleaned_data = cleaned_data.reset_index(drop=True)
    cleaned_data.index += 1
    return cleaned_data


def create_init_condition(
    thermal_units: list, storage_units: list = None
) -> dict[(str, int), dict]:
    """Return dicts of system statuses in the format {(unit, hour): value}"""
    # Assume thermal units in the systems are offline at the beginning
    initial_p = {unit_g: 0 for unit_g in thermal_units}
    initial_u = initial_p.copy()
    initial_v = initial_p.copy()
    initial_w = initial_p.copy()

    # Thermal units do not carry any minimum up and down time at the beginning
    initial_min_on = {unit_g: 0 for unit_g in thermal_units}
    initial_min_off = initial_min_on.copy()

    # Energy storage systems start with zero charge
    if storage_units is None:
        initial_charge_state = {}
    elif len(storage_units) > 0:
        initial_charge_state = {unit: 0 for unit in storage_units}
    else:
        initial_charge_state = {}

    return {
        "initial_p": initial_p,
        "initial_u": initial_u,
        "initial_v": initial_v,
        "initial_w": initial_w,
        "initial_min_on": initial_min_on,
        "initial_min_off": initial_min_off,
        "initial_charge_state": initial_charge_state,
    }


def get_node_hour_from_flow_constraint(constraint_name: str) -> tuple[str, int]:
    """Get the node and hour from the flow constraint name.

    Args:
        constraint_name: The name of the constraint.

    Returns:
        The node and hour.

    """
    flow_constraint_pattern = re.compile(r"flowBal\[(.+),(\d+)\]")
    match = flow_constraint_pattern.match(constraint_name)
    if match:
        node = match.group(1)
        hour = int(match.group(2))
        return node, hour
    else:
        return None, None


def get_unit_hour_from_varnam(var_name: str) -> tuple[str, int]:
    """Get the unit and hour from the variable name.

    Args:
        var_name: The name of the variable.

    Returns:
        The unit and hour.

    """
    node_var_pattern = re.compile(r"(\w+)\[(.+),(\d+)\]")
    match = node_var_pattern.match(var_name)
    if match:
        unit = match.group(2)
        hour = int(match.group(3))
        return unit, hour
    else:
        raise ValueError("Invalid variable name format")


def get_edge_hour_from_varname(var_name: str) -> tuple[tuple[str, str], int]:
    """Get the edge and hour from the variable name: flow[a,b,t].

    Args:
        var_name: The name of the variable.

    Returns:
        The edge and hour.

    """
    edge_var_pattern = re.compile(r"flow\[(.+),(.+),(\d+)\]")
    match = edge_var_pattern.match(var_name)
    if match:
        edge = (match.group(1), match.group(2))
        hour = int(match.group(3))
        return edge, hour
    else:
        raise ValueError("Invalid variable name format")


def get_current_time() -> str:
    return datetime.datetime.now().strftime("%Y%m%d_%H%M")


def write_df(
    df: pd.DataFrame,
    output_folder: str,
    output_name: str,
    model_id: str,
) -> None:
    """Write a dataframe to the output folder.

    Args:
        df: The dataframe to write.
        output_name: The name of the output file.
        model_id: The model ID.

    Returns:
        None
    """
    # First check that the output directory exists. If not, create it.
    if not os.path.exists(output_folder):
        os.makedirs(output_folder)
    df.to_csv(
        os.path.join(output_folder, f"{model_id}_{output_name}.csv"),
        index=False,
    )


def calc_remaining_duration(
    solution: pd.DataFrame,
    sim_horizon: int,
    thermal_units: list[str],
    duration_dict: dict[str, int],  # Generic for TU or TD
    vartype: str,  # 'startup' or 'shutdown'
) -> dict[str, int]:
    """Calculates the remaining duration (on or off) for each thermal unit.

    This function analyzes the provided solution DataFrame to determine the latest
    timestep at which a specified event (startup or shutdown) occurred for each
    thermal unit. It then calculates the remaining duration based on the simulation
    horizon and the unit's minimum required duration.

    Args:
        solution: A DataFrame containing the solution of the optimization model.
        sim_horizon: The length of the simulation horizon.
        thermal_units: A list of thermal unit names.
        duration_dict: A dictionary mapping unit names to their respective minimum durations.
        vartype: The type of event to analyze. Either 'startup' or 'shutdown'.

    Returns:
        A dictionary mapping unit names to their remaining durations.

    Raises:
        ValueError: If the simulation horizon is shorter than the maximum duration of any thermal unit.
    """

    # This logic does not work if sim_horizon is shorter than the duration
    if sim_horizon < max(duration_dict.values()):
        raise ValueError(
            "The simulation horizon is shorter than the maximum duration of the thermal units."
        )

    remaining_durations = {}

    for unit in thermal_units:
        subset = (
            solution[(solution["node"] == unit) & (solution["vartype"] == vartype)]
            .set_index("timestep")
            .drop(["vartype", "node"], axis=1)
        )

        filtered_df = subset[subset["value"] == 1]
        if len(filtered_df) > 0:
            latest_event_timestep = filtered_df.index.max()
        else:
            latest_event_timestep = -sim_horizon

        remaining_durations[unit] = max(
            0, duration_dict[unit] - (sim_horizon - latest_event_timestep) - 1
        )

    return remaining_durations


def calc_remaining_on_duration(
    solution: pd.DataFrame,
    sim_horizon: int,
    thermal_units: list[str],
    TU: dict[str, int],
) -> dict[str, int]:
    """Calculate the remaining online duration for each thermal unit.

    Args:
        solution: A DataFrame containing the solution of the optimization model.
        sim_horizon: The length of the simulation horizon.
        thermal_units: A list of thermal unit names.
        TU: A dictionary mapping unit names to their respective minimum online durations.

    Returns:
        A dictionary mapping unit names to their remaining online durations.
    """
    return calc_remaining_duration(solution, sim_horizon, thermal_units, TU, "startup")


def calc_remaining_off_duration(
    solution: pd.DataFrame,
    sim_horizon: int,
    thermal_units: list[str],
    TD: dict[str, int],
) -> dict[str, int]:
    """Calculate the remaining shutdown duration for each thermal unit.

    Args:
        solution: A DataFrame containing the solution of the optimization model.
        sim_horizon: The length of the simulation horizon.
        thermal_units: A list of thermal unit names.
        TD: A dictionary mapping unit names to their respective minimum offline durations.

    Returns:
        A dictionary mapping unit names to their remaining shutdown durations.
    """
    return calc_remaining_duration(solution, sim_horizon, thermal_units, TD, "shutdown")


def parse_node_variables(
    solution: pd.DataFrame, sim_horizon: int, step_k: int
) -> pd.DataFrame:
    """Parse the node variables from the solution DataFrame. Node variables are in the (node, t) format.
    Also, ensure binary values are rounded to 0 or 1.

    Args:
        solution: The solution DataFrame.
        sim_horizon: The length of the simulation horizon.
        step_k: The current simulation period.

    Returns:
        pd.DataFrame: The node variables DataFrame"""

    node_var_pattern = r"(\w+)\[(.+),(\d+)\]"
    current_node_vars = solution[solution["varname"].str.match(node_var_pattern)].copy()
    # Flow should not be included in the node variables
    current_node_vars = current_node_vars[current_node_vars["vartype"] != "flow"]

    current_node_vars[["node", "timestep"]] = current_node_vars["varname"].str.extract(
        node_var_pattern, expand=True
    )[[1, 2]]

    current_node_vars["timestep"] = current_node_vars["timestep"].astype(int)

    current_node_vars["hour"] = current_node_vars["timestep"] + sim_horizon * (
        step_k - 1
    )

    # Rounding binary values
    current_node_vars.loc[
        np.isclose(current_node_vars["value"], 0, atol=1e-4), "value"
    ] = 0
    current_node_vars.loc[
        np.isclose(current_node_vars["value"], 1, atol=1e-4), "value"
    ] = 1
    return current_node_vars


def parse_flow_variables(
    solution: pd.DataFrame, sim_horizon: int, step_k: int
) -> pd.DataFrame:
    """
    The flow variables are in the (node, node, t) format.

    Args:
        solution: The solution DataFrame.
        sim_horizon: The length of the simulation horizon.
        step_k: The current simulation period.

    Returns:
        pd.DataFrame: The flow variables DataFrame
    """
    flow_var_pattern = r"flow\[(.+),(.+),(\d+)\]"
    cur_flow_vars = solution[solution["varname"].str.match(flow_var_pattern)].copy()

    cur_flow_vars[["node_a", "node_b", "timestep"]] = cur_flow_vars[
        "varname"
    ].str.extract(flow_var_pattern, expand=True)

    cur_flow_vars["timestep"] = cur_flow_vars["timestep"].astype(int)
    cur_flow_vars["hour"] = cur_flow_vars["timestep"] + sim_horizon * (step_k - 1)
    cur_flow_vars = cur_flow_vars.drop("varname", axis=1)
    return cur_flow_vars


def parse_syswide_variables(
    solution: pd.DataFrame, sim_horizon: int, step_k: int
) -> pd.DataFrame:
    """
    The system-wide variables are in the (t) format.

    Args:
        solution: The solution DataFrame.
        sim_horizon: The length of the simulation horizon.
        step_k: The current simulation period.

    Returns:
        pd.DataFrame: The system-wide variables DataFrame
    """
    syswide_var_pattern = r"(.+)\[(\d+)\]"
    cur_syswide_vars = solution[
        solution["varname"].str.match(syswide_var_pattern)
    ].copy()

    cur_syswide_vars["timestep"] = cur_syswide_vars["varname"].str.extract(
        syswide_var_pattern, expand=True
    )[1]
    cur_syswide_vars["timestep"] = cur_syswide_vars["timestep"].astype(int)
    cur_syswide_vars["hour"] = cur_syswide_vars["timestep"] + sim_horizon * (step_k - 1)
    cur_syswide_vars = cur_syswide_vars.drop("varname", axis=1)
    return cur_syswide_vars


def parse_lmp(lmp: dict[str, float], sim_horizon: int, step_k: int) -> pd.DataFrame:
    """Parse the LMP dictionary and return a DataFrame.

    Args:
        lmp: The dictionary of LMP values.
        sim_horizon: The length of the simulation horizon.
        step_k: The current simulation period.

    Returns:
        pd.DataFrame: The LMP DataFrame.
    """
    lmp_df = pd.DataFrame.from_dict(lmp, orient="index", columns=["value"])
    lmp_df = lmp_df.reset_index().rename(columns={"index": "name"})
    lmp_df[["node", "timestep"]] = lmp_df["name"].str.extract(r"flowBal\[(.*),(\d+)\]")
    lmp_df["timestep"] = lmp_df["timestep"].astype(int)
    lmp_df["hour"] = lmp_df["timestep"] + sim_horizon * (step_k - 1)
    # Keep only the first 24-hours of the simulation
    lmp_df = lmp_df[lmp_df["timestep"] <= 24]
    lmp_df = lmp_df.drop(["name"], axis=1)
    return lmp_df


def get_fuel_mix_order() -> list[str]:
    """Return the order of fuel mix for plotting.
    We use this list to order the node variables dataframe.

    Returns
        list[str]: The order of fuel mix.
    -------
    """
    return pd.read_csv(
        os.path.join(get_database_dir(), "fuels.csv"),
        header=0,
    )["name"].tolist()


def get_fuel_color_map() -> dict:
    """Return a map of fuel type to its color. This is defined in the database folder."""
    fuel_color_map = (
        pd.read_csv(
            os.path.join(get_database_dir(), "fuels.csv"),
            header=0,
            usecols=["name", "color"],
        )
        .set_index("name")
        .to_dict()["color"]
    )
    return fuel_color_map


def get_lines_params() -> pd.DataFrame:
    """Return a dataframe of line parameters, located in database/transmission_params.csv"""
    return pd.read_csv(
        os.path.join(get_database_dir(), "transmission_params.csv"),
        header=0,
    )


def create_geoseries_columns(df: pd.DataFrame) -> pd.DataFrame:
    df = df.copy()
    df["geometry"] = df.apply(
        lambda row: LineString(
            [(row["source_lon"], row["source_lat"]), (row["sink_lon"], row["sink_lat"])]
        ),
        axis=1,
    )

    df["source_location"] = df.apply(
        lambda row: Point(row["source_lon"], row["source_lat"]), axis=1
    )
    df["sink_location"] = df.apply(
        lambda row: Point(row["sink_lon"], row["sink_lat"]), axis=1
    )

    # Plotting the substations requires columns to be a GeoSeries
    df["source_location"] = gpd.GeoSeries(df["source_location"])
    df["sink_location"] = gpd.GeoSeries(df["sink_location"])
    return df



""" simulation.py: Main class to run the simulation of the power system model """

import pandas as pd
from pownet.data_utils import (
    create_init_condition,
)
from .builder import ModelBuilder
from .data_processor import DataProcessor
from .input import SystemInput
from .output import OutputProcessor
from .record import SystemRecord
from .visualizer import Visualizer


class Simulator:
    """Main class to run the simulation of the power system model"""

    def __init__(
        self,
        input_folder: str,
        model_name: str,
        model_year: int,
        frequency: int = 50,
        use_spin_var: bool = True,
        dc_opf: str = "kirchhoff",
        spin_reserve_factor: float = 0.15,
        spin_reserve_mw: float = None,
        line_loss_factor: float = 0.075,
        line_capacity_factor: float = 0.9,
        load_shortfall_penalty_factor: float = 1000,
        load_curtail_penalty_factor: float = 10,
        spin_shortfall_penalty_factor: float = 1000,
    ) -> None:
        """Initialize the simulation parameters

        Args:
            input_folder (str): The folder containing the input data.
            model_name (str): The name of the model.
            model_year (int): The year of the model.
            frequency (int): The frequency of the power system model.
            use_spin_var (bool): Whether to use spinning reserve.
            dc_opf (str): The type of DC OPF to use.
            spin_reserve_factor (float): The spinning reserve factor.
            line_loss_factor (float): The line loss factor.
            line_capacity_factor (float): The line capacity factor.
            load_shortfall_penalty_factor (float): The load shortfall penalty factor.
            load_curtail_penalty_factor (float): The load curtailment penalty factor.
            spin_shortfall_penalty_factor (float): The spinning reserve shortfall penalty factor.

        Returns:
            None
        """

        self.input_folder: str = input_folder
        self.model_name: str = model_name
        self.model_year: int = model_year

        self.frequency: int = frequency
        self.use_spin_var: bool = use_spin_var
        self.dc_opf: str = dc_opf
        self.spin_reserve_factor: float = spin_reserve_factor
        self.spin_reserve_mw: float = spin_reserve_mw
        self.line_loss_factor: float = line_loss_factor
        self.line_capacity_factor: float = line_capacity_factor
        self.load_shortfall_penalty_factor: float = load_shortfall_penalty_factor
        self.load_curtail_penalty_factor: float = load_curtail_penalty_factor
        self.spin_shortfall_penalty_factor: float = spin_shortfall_penalty_factor

        # Simulation objects
        self.inputs: SystemInput = None
        self.system_record: SystemRecord = None
        self.node_variables: pd.DataFrame = pd.DataFrame()

    def run(
        self,
        sim_horizon: int,
        steps_to_run: int,
        num_sim_days: int = 365,
        to_process_inputs: bool = True,
        solver: str = "gurobi",
        log_to_console: bool = True,
        mipgap: float = 1e-3,
        timelimit: int = 600,
        num_threads: int = 0,
        find_lmp: bool = False,
    ) -> SystemRecord:
        """Run the simulation of the power system model

        Args:
            sim_horizon (int): The simulation horizon in hours.
            steps_to_run (int): The number of steps to run the simulation.
            to_process_inputs (bool): Whether to process the input data.
            solver (str): The solver to use for optimization.
            log_to_console (bool): Whether to log the optimization output to the console.
            mipgap (float): The MIP gap for the optimization.
            timelimit (int): The time limit for the optimization in seconds.
            num_threads (int): The number of threads to use for optimization.
            find_lmp (bool): Whether to find the locational marginal prices.

        Returns:
            SystemRecord: The system record object containing the simulation results.
        """

        if steps_to_run is None:
            steps_to_run = num_sim_days - (sim_horizon // 24 - 1)

        # To create files with "pownet_" prefix
        if to_process_inputs:
            data_processor = DataProcessor(
                input_folder=self.input_folder,
                model_name=self.model_name,
                year=self.model_year,
                frequency=self.frequency,
            )
            data_processor.execute_data_pipeline()

        self.inputs = SystemInput(
            input_folder=self.input_folder,
            model_name=self.model_name,
            year=self.model_year,
            sim_horizon=sim_horizon,
            num_sim_days=num_sim_days,
            use_spin_var=self.use_spin_var,
            dc_opf=self.dc_opf,
            spin_reserve_factor=self.spin_reserve_factor,
            spin_reserve_mw=self.spin_reserve_mw,
            line_loss_factor=self.line_loss_factor,
            line_capacity_factor=self.line_capacity_factor,
            load_shortfall_penalty_factor=self.load_shortfall_penalty_factor,
            load_curtail_penalty_factor=self.load_curtail_penalty_factor,
            spin_shortfall_penalty_factor=self.spin_shortfall_penalty_factor,
        )
        # Produce an error if the data is not making sense
        self.inputs.load_and_check_data()
        self.inputs.print_summary()

        ####################### Simulation
        self.system_record = SystemRecord(self.inputs)
        model_builder = ModelBuilder(self.inputs)

        # Initially, all thermal units are off. They have to be switched on from cold start
        init_conditions = create_init_condition(
            self.inputs.thermal_units, self.inputs.storage_units
        )

        for step_k in range(1, steps_to_run + 1):
            # Build or update the model
            if step_k == 1:
                power_system_model = model_builder.build(
                    step_k=step_k,
                    init_conds=init_conditions,
                )
            else:
                power_system_model = model_builder.update(
                    step_k=step_k,
                    init_conds=init_conditions,
                )
            # Optimization
            power_system_model.optimize(
                solver=solver,
                log_to_console=log_to_console,
                mipgap=mipgap,
                timelimit=timelimit,
                num_threads=num_threads,
            )
            self.system_record.keep(
                runtime=power_system_model.get_runtime(),
                objval=power_system_model.get_objval(),
                solution=power_system_model.get_solution(),
                step_k=step_k,
                lmp=power_system_model.solve_for_lmp() if find_lmp else None,
            )
            # Update the initial conditions for the next step
            init_conditions = self.system_record.get_init_conds()

    def get_node_variables(self) -> pd.DataFrame:
        """Return the node-specific variables."""
        return self.system_record.get_node_variables()

    def get_flow_variables(self) -> pd.DataFrame:
        """Return the flow variables."""
        return self.system_record.get_flow_variables()

    def get_systemwide_variables(self) -> pd.DataFrame:
        """Return the system variables"""
        return self.system_record.get_systemwide_variables()

    def write_results(self, output_folder: str) -> None:
        """Write the simulation results to files"""
        self.system_record.write_simulation_results(output_folder)

    def plot_fuelmix(self, chart_type: str, output_folder: str = None) -> None:
        """Plot the fuel mix of the power system

        Args:
            chart_type (str): The type of chart to plot. Choose between 'bar' and 'area'.
            output_folder (str): The folder to save the plot.

        Returns:
            None
        """
        if chart_type not in ["bar", "area"]:
            raise ValueError(
                "Invalid chart type. Choose between 'fuelmix' and 'fuelmix_area'."
            )

        output_processor = OutputProcessor()
        output_processor.load(self.inputs)

        node_variables = self.system_record.get_node_variables()

        visualizer = Visualizer(model_id=self.inputs.model_id)
        if chart_type == "bar":
            visualizer.plot_fuelmix_bar(
                dispatch=output_processor.get_hourly_generation(node_variables),
                demand=output_processor.get_hourly_demand(self.inputs.demand),
                output_folder=output_folder,
            )
        elif chart_type == "area":
            visualizer.plot_fuelmix_area(
                dispatch=output_processor.get_hourly_generation(node_variables),
                demand=output_processor.get_hourly_demand(self.inputs.demand),
                output_folder=output_folder,
            )

    def plot_thermal_units(self, output_folder: str = None) -> None:
        """Plot the status of the thermal units

        Args:
            output_folder (str): The folder to save the plot.

        Returns:
            None
        """
        node_variables = self.system_record.get_node_variables()

        output_processor = OutputProcessor()
        output_processor.load(self.inputs)

        visualizer = Visualizer(model_id=self.inputs.model_id)
        visualizer.plot_thermal_units(
            unit_status=output_processor.get_thermal_unit_hourly_status(node_variables),
            thermal_dispatch=output_processor.get_thermal_unit_hourly_dispatch(
                node_variables
            ),
            thermal_rated_capacity=self.inputs.thermal_rated_capacity,
            output_folder=output_folder,
        )

    def plot_lmp(self, output_folder: str = None) -> None:
        """Plot the locational marginal prices

        Args:
            output_folder (str): The folder to save the plot.

        Returns:
            None
        """
        visualizer = Visualizer(model_id=self.inputs.model_id)
        visualizer.plot_lmp(
            lmp_df=self.system_record.get_lmp(),
            output_folder=output_folder,
        )

    # def reoperate(
    #     self,
    #     k: int,
    #     builder: ModelBuilder,
    #     init_conds: dict,
    #     mip_gap: float = None,
    #     timelimit: float = None,
    # ):
    #     raise NotImplementedError("Reoperation is not implemented yet.")
    #     reop_converge = False
    #     reop_k = 0
    #     while not reop_converge:
    #         print(f"\nReservoirs reoperation iteration {reop_k}")
    #         print("New Capacity vs. Current Dispatch")

    #         # PowNet returns the hydropower dispatch in hourly resolution across the simulation horizon
    #         hydro_dispatch, start_day, end_day = get_hydro_from_model(
    #             self.model.model, k
    #         )
    #         # Convert to daily dispatch
    #         hydro_dispatch = convert_to_daily_hydro(hydro_dispatch, start_day, end_day)
    #         new_hydro_capacity = self.reservoir_operator.reoperate_basins(
    #             pownet_dispatch=hydro_dispatch
    #         )

    #         for res in new_hydro_capacity.columns:
    #             print(
    #                 f"{res}: {round(new_hydro_capacity[res].sum(),2)} vs {round(hydro_dispatch[res].sum(),2)}",
    #             )

    #         max_deviation = (new_hydro_capacity - hydro_dispatch).abs().max()
    #         # The tolerance for convergence should be 5% of the largest hydro capacity
    #         reop_tol = 0.05 * new_hydro_capacity.max()
    #         if (max_deviation <= reop_tol[max_deviation.index]).all():
    #             reop_converge = True
    #             print(f"PowNet: Day {k+1} - Reservoirs converged at iteration {reop_k}")

    #         if reop_k > 50:
    #             raise ValueError(
    #                 "Reservoirs reoperation did not converge after 100 iterations"
    #             )

    #         # To reoptimize PowNet with the new hydropower capacity,
    #         # update the builder class
    #         builder.update_hydro_capacity(new_hydro_capacity)
    #         self.model = builder.update(
    #             k=k,
    #             init_conds=init_conds,
    #             mip_gap=mip_gap,
    #             timelimit=timelimit,
    #         )
    #         self.model.optimize()

    #         # Keep track of optimization time oand reoperation iterations
    #         self.reop_opt_time += self.model.get_runtime()
    #         reop_k += 1

    #     # Record the number of iterations after convergence
    #     self.reop_iter.append(reop_k)



""" This is the core module.
"""

from .builder import ModelBuilder
from .input import SystemInput
from .output import OutputProcessor
from .visualizer import Visualizer
from .record import SystemRecord
from .simulation import Simulator
from .data_processor import DataProcessor



""" demand.py: Model for solar time series data"""

import numpy as np
import pandas as pd
from pmdarima import auto_arima
from statsmodels.regression.linear_model import OLS
from statsmodels.tsa.statespace.sarimax import SARIMAX
from statsmodels.tsa.seasonal import STL, DecomposeResult

from .timeseries_model import TimeSeriesModel

import logging

logger = logging.getLogger(__name__)


class DemandTSModel(TimeSeriesModel):
    def __init__(
        self,
    ) -> None:
        super().__init__()
        self._monthly_models: dict[int, SARIMAX] = {}
        self._predictions: pd.Series = pd.Series()
        self._pred_residuals: pd.Series = pd.Series()

        # Specific model parameters
        self.monthly_reg_models: dict[int, OLS] = {}
        self.stl_seasonal_value = 24 * 7 - 1  # Must be an odd number
        self.stl_period_value = 24
        self.monthly_stl_results: dict[int, DecomposeResult] = {}

    @property
    def monthly_models(self) -> dict:
        if not self._is_fitted:
            raise ValueError("Model must be fitted first!")
        return self._monthly_models

    @property
    def predictions(self) -> pd.Series:
        if not self._is_fitted:
            raise ValueError("Model must be fitted first!")
        return self._predictions

    @property
    def pred_residuals(self) -> pd.Series:
        if not self._is_fitted:
            raise ValueError("Model must be fitted first!")
        return self._pred_residuals

    def _fit(
        self,
        target_column: str,
        exog_vars: list[str],
        arima_order: tuple[int, int, int],
        seasonal_order: tuple[int, int, int, int] = None,
    ) -> None:
        """Steps include
        1) Fit a regression to get yt
        2) Fit an STL to the regression residuals to get ytt
        3) Fit a SARIMAX model to ytt

        Note that the user must at lease use temperature as a predictor
        """
        self._pred_residuals = pd.Series()
        for month in self.months:
            logger.info(f"Fitting SARIMAX model for month {month}")
            # Subset the data for the month
            monthly_y = self.data.loc[self.data.index.month == month, target_column]
            monthly_exog = self.data.loc[
                self.data.index.month == month, exog_vars
            ].astype(float)

            # Regression model
            reg_model = OLS(monthly_y, monthly_exog).fit()
            monthly_yt = reg_model.resid

            # STL model
            stl_model = STL(
                monthly_yt,
                seasonal=self.stl_seasonal_value,
                period=self.stl_period_value,
            )
            stl_result = stl_model.fit()

            # SARIMAX model
            monthly_ytt = monthly_yt - stl_result.seasonal - stl_result.trend
            sarimax_model = SARIMAX(
                monthly_ytt,
                order=arima_order,
                seasonal_order=seasonal_order,
            ).fit(disp=True)

            # Store the models, and residuals
            self.monthly_reg_models[month] = reg_model
            self._monthly_models[month] = sarimax_model
            self.monthly_stl_results[month] = stl_result

            if self._pred_residuals.empty:
                self._pred_residuals = sarimax_model.resid
            else:
                self._pred_residuals = pd.concat(
                    [self._pred_residuals, sarimax_model.resid]
                )
        self._pred_residuals.name = "value"

    def _predict(self) -> pd.Series:
        self._predictions = pd.Series()
        for month in self.months:
            monthly_exog = self.data.loc[
                self.data.index.month == month, self.exog_vars
            ].astype(float)
            sarimax_model = self._monthly_models[month]
            # SARIMAX model ytt, which is the trend and residuals of LOESS
            monthly_ytt_pred = sarimax_model.predict()
            monthly_yt_pred = (
                monthly_ytt_pred + self.monthly_stl_results[month].seasonal
            )

            # Regression model to get y
            reg_pred = self.monthly_reg_models[month].predict(monthly_exog)
            monthly_y_pred = monthly_yt_pred + reg_pred

            monthly_y_pred.name = "value"
            if self._predictions.empty:
                self._predictions = monthly_y_pred
            else:
                self._predictions = pd.concat([self._predictions, monthly_y_pred])
        return self._predictions

    def _get_synthetic(self, exog_data: pd.DataFrame, seed: int) -> pd.Series:
        synthetic_y = pd.Series()
        for month in self.months:
            # Models are fitted for each month
            stl_result = self.monthly_stl_results[month]
            sarimax_model = self._monthly_models[month]

            monthly_exog = exog_data.loc[exog_data.index.month == month]

            # Create a synthetic time series for each day
            for day in monthly_exog.index.day.unique():
                # Regression model
                start_time = monthly_exog.loc[monthly_exog.index.day == day].index[0]
                end_time = start_time + pd.Timedelta(hours=23)
                daily_exog = monthly_exog.loc[start_time:end_time]
                reg_pred = self.monthly_reg_models[month].predict(exog=daily_exog)
                reg_pred = pd.Series(reg_pred, index=daily_exog.index)

                # Bootstrap the SARIMAX predictions within 95% confidence interval
                sarimax_pred = sarimax_model.predict()
                sarimax_pred_ci = np.percentile(sarimax_pred, [2.5, 97.5])
                sarimax_bootstrap = np.random.choice(
                    sarimax_pred_ci, size=len(daily_exog)
                )
                sarimax_bootstrap = pd.Series(sarimax_bootstrap, index=daily_exog.index)

                # Recover electricity demand by adding predictions from the regression model,
                # predictions from SARIMAX, and also the STL's seasonal and trend components
                daily_stl_seasonal = stl_result.seasonal.loc[start_time:end_time]
                daily_stl_trend = stl_result.trend.loc[start_time:end_time]
                daily_syn_y = (
                    sarimax_bootstrap + reg_pred + daily_stl_seasonal + daily_stl_trend
                )
                daily_syn_y.name = "value"
                if synthetic_y.empty:
                    synthetic_y = daily_syn_y
                else:
                    synthetic_y = pd.concat([synthetic_y, daily_syn_y])
        return synthetic_y

    def _find_best_model(
        self,
        target_column: str,
        exog_vars: list[str],
        month_to_use: int,
        seed: int,
        suppress_warnings: bool,
    ) -> tuple[tuple[int, int, int], tuple[int, int, int, int]]:
        monthly_y = self.data.loc[self.data.index.month == month_to_use, target_column]
        monthly_exog = self.data.loc[
            self.data.index.month == month_to_use, exog_vars
        ].astype(float)

        # Regression model
        reg_model = OLS(monthly_y, monthly_exog).fit()
        monthly_yt = reg_model.resid

        # STL model
        stl_model = STL(
            monthly_yt,
            seasonal=self.stl_seasonal_value,
            period=self.stl_period_value,
        )
        stl_result = stl_model.fit()

        # SARIMAX model
        monthly_ytt = monthly_yt - stl_result.seasonal - stl_result.trend
        best_model = auto_arima(
            monthly_ytt,
            X=monthly_exog,
            start_p=0,
            start_q=0,
            max_p=2,
            max_d=2,
            max_q=2,
            seasonal=False,
            information_criterion="aic",
            stepwise=True,
            suppress_warnings=suppress_warnings,
            error_action="warn",
            random_state=seed,
        )
        return best_model.order, best_model.seasonal_order



""" input.py: SystemInput class loads and checks the input data. It is used by other PowNet objects to access the input data. """

from datetime import datetime
import logging
import json
import os
import textwrap

import gurobipy as gp
from gurobipy import GRB
import pandas as pd

logger = logging.getLogger(__name__)


class SystemInput:
    def __init__(
        self,
        input_folder: str,
        model_name: str,
        year: int,
        sim_horizon: int,
        num_sim_days: int = 365,
        use_spin_var: bool = True,
        dc_opf: str = "kirchhoff",
        spin_reserve_factor: float = 0.15,
        spin_reserve_mw: float = None,
        line_loss_factor: float = 0.075,
        line_capacity_factor: float = 0.9,
        load_shortfall_penalty_factor: float = 1000,
        load_curtail_penalty_factor: float = 900,
        spin_shortfall_penalty_factor: float = 1000,
    ) -> None:
        """This class reads the input data for the power system model."""

        self.model_name: str = model_name
        self.model_dir: str = os.path.join(input_folder, model_name)
        self.year: int = year
        self.sim_horizon: int = sim_horizon

        self.num_sim_days: int = num_sim_days
        self.num_sim_hours: int = num_sim_days * 24

        self.use_spin_var: bool = use_spin_var

        # The timestamp is used to create a unique folder for the model
        self.timestamp: str = datetime.now().strftime("%Y%m%d_%H%M")

        self.model_id: str = f"{self.timestamp}_{self.model_name}_{self.sim_horizon}"

        # DC representation of the power flow model
        if dc_opf not in ["kirchhoff", "voltage_angle"]:
            raise ValueError(
                "PowNet: Line flow must be either 'kirchhoff' or 'voltage_angle'."
            )
        self.dc_opf: str = dc_opf

        # The spin reserve factor is a fraction of total demand, so should be between 0 and 1
        if not 0 <= spin_reserve_factor <= 1:
            raise ValueError("PowNet: Spin reserve factor must be between 0 and 1.")
        self.spin_reserve_factor: float = spin_reserve_factor
        self.spin_reserve_mw: float = spin_reserve_mw

        # The line loss factor is also called the inefficiency factor.
        # It is the fraction of power lost in the transmission line.
        self.line_loss_factor: float = line_loss_factor

        # The line capacity factor is the fraction of the line capacity
        # that can be used. It is used to account for the uncertainty in
        # the line capacity.
        self.line_capacity_factor: float = line_capacity_factor

        # The shortfall penalty is the cost of not meeting the demand. (USD/MWh)
        self.load_shortfall_penalty_factor: float = load_shortfall_penalty_factor
        self.load_curtail_penalty_factor: float = load_curtail_penalty_factor

        # The reserve penalty is the cost of not meeting the reserve requirement. (USD/MWh)
        self.spin_shortfall_penalty_factor: float = spin_shortfall_penalty_factor

        #################
        # Complex attributes that will be defined in load_data
        #################

        # Thermal units
        self.thermal_unit_node: dict[str, str] = {}

        self.thermal_fixed_cost: dict[str, float] = {}
        self.thermal_opex: dict[str, float] = {}
        self.thermal_startup_cost: dict[str, float] = {}
        self.thermal_heat_rate: dict[str, float] = {}

        self.thermal_rated_capacity: dict[str, float] = {}
        self.thermal_derated_capacity: pd.DataFrame = pd.DataFrame()
        self.thermal_min_capacity: dict[str, float] = {}

        self.TD: dict[str, int] = {}
        self.TU: dict[str, int] = {}
        self.SD: dict[str, float] = {}
        self.SU: dict[str, float] = {}
        self.RD: dict[str, float] = {}
        self.RU: dict[str, float] = {}

        # Energy storage
        self.ess_unit_attach: dict[str, str] = {}
        self.ess_attach_unit: dict[str, list[str]] = {}

        self.ess_hydro_units: dict[str, list[str]] = {}
        self.ess_daily_hydro_units: dict[str, list[str]] = {}
        self.ess_solar_units: dict[str, list[str]] = {}
        self.ess_wind_units: dict[str, list[str]] = {}
        self.ess_thermal_units: dict[str, list[str]] = {}
        self.ess_substation_units: dict[str, list[str]] = {}

        self.ess_max_charge: dict[str, float] = {}
        self.ess_max_discharge: dict[str, float] = {}
        self.ess_max_capacity: dict[str, float] = {}
        self.ess_min_capacity: dict[str, float] = {}
        self.ess_charge_efficiency: dict[str, float] = {}
        self.ess_discharge_efficiency: dict[str, float] = {}
        self.ess_self_discharge_rate: dict[str, float] = {}
        self.ess_derated_capacity: pd.DataFrame = pd.DataFrame()

        self.ess_unit_node: dict[str, str] = {}

        # Demand
        self.demand: pd.DataFrame = pd.DataFrame()
        self.demand_nodes: list[str] = []
        self.max_demand_node: str = ""

        # Hydropower (hourly and daily timeseries)
        self.hydro_contracted_capacity: dict[str, float] = {}
        self.hydro_capacity: pd.DataFrame = pd.DataFrame()
        self.hydro_unit_node: dict[str, str] = {}
        self.daily_hydro_capacity: pd.DataFrame = pd.DataFrame()
        self.daily_hydro_unit_node: dict[str, str] = {}

        self.solar_contracted_capacity: dict[str, float] = {}
        self.solar_capacity: pd.DataFrame = pd.DataFrame()
        self.solar_unit_node: dict[str, str] = {}

        self.wind_contracted_capacity: dict[str, float] = {}
        self.wind_capacity: pd.DataFrame = pd.DataFrame()
        self.wind_unit_node: dict[str, str] = {}

        self.import_contracted_capacity: dict[str, float] = {}
        self.import_capacity: pd.DataFrame = pd.DataFrame()
        self.import_unit_node: dict[str, str] = {}

        self.edges: gp.tuplelist = gp.tuplelist()
        self.line_capacity: pd.DataFrame = pd.DataFrame()
        self.rated_line_capacities: dict[tuple[str, str], int] = {}
        self.line_locations: pd.DataFrame = pd.DataFrame()
        self.susceptance: pd.DataFrame = pd.DataFrame()
        self.cycle_map: dict = {}

        self.max_line_capacity: int = 0
        self.spin_requirement: pd.Series = pd.Series()

        self.fuelmap: dict[str, str] = {}  # Maps each unit to its fuel type

        # Maps a unit to its contracted price
        self.fuel_contracts: dict[str, str] = {}
        self.nondispatch_contracts: dict[str, str] = {}
        self.ess_contracts: dict[str, str] = {}
        # Contract costs are dicts of (contract, timestep) -> cost_per_mw
        self.contract_costs: dict[tuple[str, int], float] = {}

        # List of units
        self.thermal_units: list[str] = []
        self.thermal_must_take_units: list[str] = []

        self.hydro_units: list[str] = []
        self.solar_units: list[str] = []
        self.wind_units: list[str] = []
        self.import_units: list[str] = []
        self.all_generators: set[str] = []

        self.storage_units: list = []

        self.hydro_must_take_units: list[str] = []
        self.daily_hydro_must_take_units: list[str] = []
        self.solar_must_take_units: list[str] = []
        self.wind_must_take_units: list[str] = []
        self.import_must_take_units: list[str] = []

        # Generators by node
        self.node_generator: dict[str, list[str]] = {}

        # Edges by node
        self.nodes: set[str] = set(["b1"])  # Will get overwritten by the actual nodes
        self.node_edge: dict[str, list[str]] = {}

    def _load_timeseries_from_csv(
        self, filename: str, header_levels: int
    ) -> pd.DataFrame:
        """Helper function to load CSV with default options.
        - Date columns are dropped from the DataFrame
        - PowNet indexing starts at 1
        """
        date_cols = ["year", "month", "day", "hour", "date", "datetime"]
        # If there are header levels, we drop the date columns at the lowest level
        col_level = None
        if header_levels > 0:
            col_level = 0
        timeseries = pd.read_csv(
            os.path.join(self.model_dir, filename),
            header=list(range(header_levels + 1)),
        ).drop(date_cols, level=col_level, axis=1, errors="ignore")
        timeseries.index += 1
        return timeseries

    def _check_and_load_timeseries(
        self, filename: str, header_levels: int
    ) -> pd.DataFrame:
        """Check if the timeseries file exists and load it.
        Timeseries of unit capacities are column indexed with unit name and the connected node.
        """
        if os.path.exists(os.path.join(self.model_dir, filename)):
            return self._load_timeseries_from_csv(filename, header_levels=header_levels)
        return pd.DataFrame()

    def _check_and_load_csv(self, filename: str) -> pd.DataFrame:
        """Check if the CSV file exists and load it."""
        if os.path.exists(os.path.join(self.model_dir, filename)):
            return pd.read_csv(os.path.join(self.model_dir, filename), header=0)
        return pd.DataFrame()

    def _get_column_pairs_as_dict(self, df: pd.DataFrame) -> list:
        """
        Extracts dictionary pairs from the DataFrame's multi-index columns.

        Args:
            df: The DataFrame with multi-index columns.

        Returns:
            A dictionary where (1) Keys The first-level column index values and (2) Values are the second-level column index values.
        """
        column_pairs = df.columns.to_flat_index().tolist()
        return dict(column_pairs)

    def load_thermal_unit_params(self):
        """Load the techno-economic parameters of thermal units from thermal_unit.csv."""
        thermal_unit_df = self._check_and_load_csv("thermal_unit.csv")

        if thermal_unit_df.empty:
            return

        thermal_unit_df = thermal_unit_df.set_index("name")

        self.thermal_unit_node = thermal_unit_df["node"].to_dict()
        self.thermal_fixed_cost = thermal_unit_df["fixed_cost"].to_dict()
        self.thermal_opex = thermal_unit_df["operation_cost"].to_dict()
        self.thermal_startup_cost = thermal_unit_df["startup_cost"].to_dict()
        self.thermal_heat_rate = thermal_unit_df["heat_rate"].to_dict()

        # The minimum downtime and uptime are in hours
        self.TD = thermal_unit_df["min_downtime"].to_dict()
        self.TU = thermal_unit_df["min_uptime"].to_dict()

        # The startup and shutdown ramp rates are in MW/h
        # IMPORTANT: Currently we do not model these values in PowNet
        # Therefore, we assume the shutdown and startup ramp rates are equal
        # and they are equal to the minimum capacity of the unit.
        self.SD = thermal_unit_df["min_capacity"].to_dict()
        self.SU = self.SD.copy()

        # The ramp rates are in MW/h.
        # IMPORTANT: We assume the ramp rates are equal for both directions.
        self.RD = thermal_unit_df["ramp_rate"].to_dict()
        self.RU = self.RD.copy()

        # Maximum/Minimum capacity of the thermal units
        self.thermal_rated_capacity = thermal_unit_df["max_capacity"].to_dict()
        self.thermal_min_capacity = thermal_unit_df["min_capacity"].to_dict()

        # The maximum capacity is reduced by the derating factor (timeseries)
        self.thermal_derated_capacity = self._load_timeseries_from_csv(
            "pownet_thermal_derated_capacity.csv", header_levels=0
        )

        # The fuel type of each thermal unit
        self.fuelmap.update(thermal_unit_df["fuel_type"].to_dict())

        # Must take thermal units (units that have dispatch priority)
        self.thermal_must_take_units = (
            thermal_unit_df.loc[thermal_unit_df["must_take"] == 1]
        ).index.tolist()

    def load_ess_params(self):
        """Load the techno-economic parameters of energy storage systems from energy_storage.csv."""
        ess_df = self._check_and_load_csv("energy_storage.csv")

        if ess_df.empty:
            return

        # Ensure values are between zero and one
        for col in ["charge_efficiency", "discharge_efficiency", "self_discharge_rate"]:
            if not 0 <= ess_df[col].min() <= ess_df[col].max() <= 1:
                raise ValueError(f"PowNet: {col} must be between 0 and 1.")

        # Min capacity must be less than max capacity
        if not (ess_df["min_capacity"] <= ess_df["max_capacity"]).all():
            raise ValueError(
                "PowNet: Min capacity must be less than or equal to max capacity."
            )

        ess_df = ess_df.set_index("name")

        self.ess_max_charge = ess_df["max_charge"].to_dict()
        self.ess_max_discharge = ess_df["max_discharge"].to_dict()
        self.ess_max_capacity = ess_df["max_capacity"].to_dict()
        self.ess_min_capacity = ess_df["min_capacity"].to_dict()
        self.ess_charge_efficiency = ess_df["charge_efficiency"].to_dict()
        self.ess_discharge_efficiency = ess_df["discharge_efficiency"].to_dict()
        self.ess_self_discharge_rate = ess_df["self_discharge_rate"].to_dict()

        # ESS -> unit/node attached
        self.ess_unit_attach = ess_df["attach_to"].to_dict()
        # unit/node -> list of ESS attached
        for unit, attached in self.ess_unit_attach.items():
            if attached not in self.ess_attach_unit.keys():
                self.ess_attach_unit[attached] = [unit]
            else:
                self.ess_attach_unit[attached].append(unit)

            # Separate ESS by unit type
            if attached in self.hydro_unit_node.keys():
                self.ess_hydro_units[attached] = self.ess_attach_unit[attached]

            elif attached in self.daily_hydro_unit_node.keys():
                self.ess_daily_hydro_units[attached] = self.ess_attach_unit[attached]

            elif attached in self.solar_unit_node.keys():
                self.ess_solar_units[attached] = self.ess_attach_unit[attached]

            elif attached in self.wind_unit_node.keys():
                self.ess_wind_units[attached] = self.ess_attach_unit[attached]

            elif attached in self.thermal_unit_node.keys():
                self.ess_thermal_units[attached] = self.ess_attach_unit[attached]

            # Lowest priority so a storage unit is chosen as attached to a generator first
            elif attached in self.nodes:
                self.ess_substation_units[attached] = self.ess_attach_unit[attached]

        self.storage_units = list(self.ess_unit_attach.keys())

        # The maximum capacity is reduced by the derating factor (timeseries)
        self.ess_derated_capacity = self._load_timeseries_from_csv(
            "pownet_ess_derated_capacity.csv", header_levels=0
        )

    def _add_units_to_node(self, node, unit_node_dict):
        """
        Helper method to add units to the node_generator dictionary.
        """
        for unit, unit_node in unit_node_dict.items():
            if unit_node == node:
                self.node_generator[node].append(unit)

    def _store_generators_by_node(self):
        """
        Store the generators by node in a dictionary : {
        node1:['generator1', 'gen2'],
        node2:['gen3', 'gen4']
        }
        Generators include thermal, hydro, solar, wind, and import units.
        """
        self.node_generator = {}
        for node in self.nodes:
            self.node_generator[node] = []
            self._add_units_to_node(node, self.thermal_unit_node)
            self._add_units_to_node(node, self.hydro_unit_node)
            self._add_units_to_node(node, self.daily_hydro_unit_node)
            self._add_units_to_node(node, self.solar_unit_node)
            self._add_units_to_node(node, self.wind_unit_node)
            self._add_units_to_node(node, self.import_unit_node)

    def _store_edges_by_node(self):
        """
        Store the edges by node in a dictionary : {
        node1:['edge1', 'edge2'],
        node2:['edge3', 'edge4']
        }
        """
        self.node_edge = {}
        for node in self.nodes:
            self.node_edge[node] = []
            for edge in self.edges:
                if node in edge:
                    self.node_edge[node].append(edge)

    def _load_capacity_and_update_fuelmap_and_get_unit_node(
        self, csv_filename: str, fuel_type: str
    ) -> tuple[pd.DataFrame, dict[str, str]]:
        """
        Loads capacity data from a CSV file, extracts unit-node mapping, and updates the fuelmap.

        Args:
            csv_filename: The name of the CSV file to load.
            fuel_type: The fuel type to associate with the loaded units.

        Returns:
            The loaded capacity DataFrame with a single-level column index.
        """
        capacity_data = self._check_and_load_timeseries(csv_filename, header_levels=1)
        unit_node_map = self._get_column_pairs_as_dict(capacity_data)
        capacity_data.columns = (
            capacity_data.columns.droplevel(1)
            if len(capacity_data.columns) >= 1
            else capacity_data.columns
        )
        self.fuelmap.update({k: fuel_type for k in unit_node_map.keys()})
        return capacity_data, unit_node_map

    def _load_contract_costs(self):
        self.fuel_contracts = self._check_and_load_csv("thermal_unit.csv")
        if not self.fuel_contracts.empty:
            self.fuel_contracts = self.fuel_contracts.set_index("name")[
                "fuel_contract"
            ].to_dict()

        self.nondispatch_contracts = self._check_and_load_csv("nondispatch_unit.csv")
        if not self.nondispatch_contracts.empty:
            self.nondispatch_contracts = self.nondispatch_contracts.set_index("name")[
                "contract"
            ].to_dict()

        # Add ESS contracts to nondispatch_contracts
        self.ess_contracts = self._check_and_load_csv("energy_storage.csv")
        if not self.ess_contracts.empty:
            self.ess_contracts = self.ess_contracts.set_index("name")[
                "cost_contract"
            ].to_dict()

        # Load contract costs as a dictionary containing (contract, timestep) -> cost_per_mw
        contract_costs_df = self._check_and_load_timeseries(
            "contract_cost.csv", header_levels=0
        )

        # Check that the contract costs timeseries is of length num_sim_hours
        if len(contract_costs_df) not in [0, self.num_sim_hours]:
            raise ValueError(
                f"PowNet: Marginal cost timeseries must be of length {self.num_sim_hours}."
            )

        self.contract_costs = {
            (col, idx): value
            for col in contract_costs_df.columns
            for idx, value in contract_costs_df[col].items()
        }

    def _load_hydropower(self) -> None:

        # Units with hourly timeseries
        if os.path.exists(os.path.join(self.model_dir, "hydropower.csv")):
            self.hydro_capacity, self.hydro_unit_node = (
                self._load_capacity_and_update_fuelmap_and_get_unit_node(
                    "hydropower.csv", fuel_type="hydropower"
                )
            )

        # Units with daily timeseries
        if os.path.exists(os.path.join(self.model_dir, "hydropower_daily.csv")):
            self.daily_hydro_capacity, self.daily_hydro_unit_node = (
                self._load_capacity_and_update_fuelmap_and_get_unit_node(
                    "hydropower_daily.csv", fuel_type="hydropower"
                )
            )

        # Check that the names do not repeat across different types
        repeated_units = set(self.hydro_unit_node.keys()).intersection(
            self.daily_hydro_unit_node.keys()
        )
        if repeated_units:
            raise ValueError(
                f"PowNet: Found hydropower units to formulate with both hourly and daily formulations: {repeated_units}"
            )

    def _load_nondispatchable_must_take_units(self):
        # A system can comprise only thermal units
        if not os.path.exists(os.path.join(self.model_dir, "nondispatch_unit.csv")):
            return

        must_take_info = pd.read_csv(
            os.path.join(self.model_dir, "nondispatch_unit.csv"), header=0
        ).set_index("name")["must_take"]

        unit_types = {
            "hydro": self.hydro_unit_node,
            "daily_hydro": self.daily_hydro_unit_node,
            "solar": self.solar_unit_node,
            "wind": self.wind_unit_node,
            "import": self.import_unit_node,
        }

        must_take_units = {
            unit_type: [unit for unit in units if must_take_info.get(unit) == 1]
            for unit_type, units in unit_types.items()
        }
        for unit_type in unit_types:
            setattr(
                self,
                f"{unit_type}_must_take_units",
                must_take_units.get(
                    unit_type, []
                ),  # Return an empty list if the unit type is not present
            )

    def _create_timeseries_of_edges(self, data: list, column_name: str) -> pd.DataFrame:
        """
        Creates a time series DataFrame from given data.

        Args:
            data: The data to be used for the time series.
            column_name: Name of the data or what it represents

        Returns:
            A DataFrame with the time series data.
        """
        df = pd.DataFrame(
            data,
            index=pd.MultiIndex.from_tuples(self.edges, names=["source", "sink"]),
            columns=[column_name],
        ).T
        # Repeat values for every hour of the year
        df = df.loc[df.index.repeat(self.num_sim_hours)].reset_index(drop=True)
        df.index += 1
        return df

    def _load_contracted_capacity(self) -> None:
        nondispatch_df = self._check_and_load_csv("nondispatch_unit.csv")
        if nondispatch_df.empty:
            return
        # Replace -1 in the contracted_capacity column to GRB.INFINITY
        nondispatch_df["contracted_capacity"] = (
            nondispatch_df["contracted_capacity"]
            .astype(float)
            .replace(-1, GRB.INFINITY)
        )

        unit_types = {
            "hydro": self.hydro_unit_node,
            "solar": self.solar_unit_node,
            "wind": self.wind_unit_node,
            "import": self.import_unit_node,
        }
        for unit_type, units in unit_types.items():
            setattr(
                self,
                f"{unit_type}_contracted_capacity",
                nondispatch_df.loc[nondispatch_df["name"].isin(units.keys())]
                .set_index("name")["contracted_capacity"]
                .to_dict(),
            )

        # Add daily hydro units to the contracted capacity under "hydro"
        self.hydro_contracted_capacity.update(
            nondispatch_df.loc[nondispatch_df["name"].isin(self.daily_hydro_unit_node)]
            .set_index("name")["contracted_capacity"]
            .to_dict()
        )

    def load_data(self):
        """Load the input data for the power system model.
        Timeseries are loaded as dataframes with the index starting at 1.
        """

        #################
        # Thermal units
        #################
        self.load_thermal_unit_params()

        #################
        # Demand (timeseries)
        #################

        self.demand = self._load_timeseries_from_csv(
            "demand_export.csv", header_levels=0
        )

        # Demand nodes
        self.demand_nodes = self.demand.columns.tolist()
        # Identify the node with the maximum demand
        self.max_demand_node = self.demand.idxmax().idxmax()

        #################
        # Hydropower
        #################

        self._load_hydropower()

        #################
        # Renewables (timeseries)
        #################

        self.solar_capacity, self.solar_unit_node = (
            self._load_capacity_and_update_fuelmap_and_get_unit_node(
                "solar.csv", "solar"
            )
        )
        self.wind_capacity, self.wind_unit_node = (
            self._load_capacity_and_update_fuelmap_and_get_unit_node("wind.csv", "wind")
        )
        self.import_capacity, self.import_unit_node = (
            self._load_capacity_and_update_fuelmap_and_get_unit_node(
                "import.csv", "import"
            )
        )

        self._load_contracted_capacity()

        #################
        # All generators
        #################
        generators = (
            list(self.thermal_unit_node.keys())
            + list(self.hydro_unit_node.keys())
            + list(self.daily_hydro_unit_node.keys())
            + list(self.solar_unit_node.keys())
            + list(self.wind_unit_node.keys())
            + list(self.import_unit_node.keys())
        )
        set_generators = set(generators)
        if len(generators) != len(set_generators):
            raise ValueError(
                "PowNet: Generator names cannot repeat across different types."
            )
        self.all_generators = set_generators

        #################
        # Must-take non-dispatchable units
        # hydro, solar, wind, import
        #################
        self._load_nondispatchable_must_take_units()

        #################
        # Transmission
        #################

        transmission: pd.DataFrame = self._check_and_load_csv("pownet_transmission.csv")

        # Nodes are connected by transmission lines
        if not transmission.empty:
            self.nodes = set(transmission.source).union(set(transmission.sink))

            # Edges are the transmission lines
            self.edges = gp.tuplelist(
                transmission.set_index(["source", "sink"]).index.tolist()
            )

            # The line susceptance is a function of hydroclimatic conditions (temperature),
            # so it is best being a timeseries. Line susceptance is used when the line flow
            # is modeled with voltage angle.
            self.susceptance = self._create_timeseries_of_edges(
                data=transmission.susceptance.values, column_name="susceptance"
            )

            # Similar to susceptance, the line capacity is also a timeseries. We use line capacity when
            # the line flow is modeled with Kirchhoff's laws.
            self.line_capacity = self._create_timeseries_of_edges(
                data=transmission.line_capacity.values, column_name="line_capacity"
            )

            self.max_line_capacity = self.line_capacity.max().max()

            # Capacity of each line segment
            self.rated_line_capacities = (
                transmission[["source", "sink", "line_capacity"]]
                .set_index(["source", "sink"])
                .to_dict()["line_capacity"]
            )

            # Source sink locations of the line segments
            try:
                self.line_locations = pd.read_csv(
                    os.path.join(self.model_dir, "transmission.csv"),
                    header=0,
                    usecols=[
                        "source",
                        "sink",
                        "source_lon",
                        "source_lat",
                        "sink_lon",
                        "sink_lat",
                    ],
                ).set_index(["source", "sink"])

            except ValueError as e:
                error_msg = e.args[0]
                if not ("columns expected but not found" in error_msg):
                    raise ValueError(f"Unexpected error: {error_msg}")

            # A user can use DataProcessor to generate pownet_cycle_map.json
            if self.dc_opf == "kirchhoff":
                if os.path.exists(
                    os.path.join(self.model_dir, "pownet_cycle_map.json")
                ):
                    with open(
                        os.path.join(self.model_dir, "pownet_cycle_map.json")
                    ) as f:
                        self.cycle_map = json.load(f)

        #################
        # System requirements
        #################
        if self.spin_reserve_mw is not None:
            self.spin_requirement = pd.Series(
                self.spin_reserve_mw, index=range(1, self.num_sim_hours + 1)
            )
        else:
            self.spin_requirement = self.demand.sum(axis=1) * self.spin_reserve_factor

        #################
        # List of units
        #################
        self.thermal_units = list(self.thermal_unit_node.keys())
        self.hydro_units = list(self.hydro_unit_node.keys()) + list(
            self.daily_hydro_unit_node.keys()
        )
        self.solar_units = list(self.solar_unit_node.keys())
        self.wind_units = list(self.wind_unit_node.keys())
        self.import_units = list(self.import_unit_node.keys())

        #################
        # Node information
        #################
        self._store_generators_by_node()
        self._store_edges_by_node()

        #################
        # Energy storage is loaded last because
        # it is linked to generators and nodes
        #################
        self.load_ess_params()

        #################
        # Contract costs
        #################
        self._load_contract_costs()

    def check_data(self):
        """
        Perform checks on the input data to ensure consistency and correctness.
        """
        ##################################
        # Simulation horizon is a multiple of 24 and greater than 24
        ##################################

        if self.sim_horizon < 24 or self.sim_horizon % 24 != 0:
            raise ValueError(
                "PowNet: Simulation horizon must be a multiple of 24 and greater than 24."
            )

        ##################################
        # Nodes are connected to the grid
        ##################################
        if not set(self.demand_nodes).issubset(self.nodes):
            raise ValueError(
                f"PowNet: Demand nodes must be connected to the grid: {set(self.demand_nodes) - self.nodes}"
            )

        nodes_to_check = [
            ("hydro_unit_node", "Hydropower units"),
            ("daily_hydro_unit_node", "Daily hydropower units"),
            ("solar_unit_node", "Solar units"),
            ("wind_unit_node", "Wind units"),
            ("import_unit_node", "Import units"),
        ]

        for nodes, node_type in nodes_to_check:
            if not set(getattr(self, nodes).values()).issubset(self.nodes):
                raise ValueError(
                    f"PowNet: {node_type} must be connected to the grid: {set(getattr(self, nodes)) - self.nodes}"
                )

        ##################################
        # Factors are between 0 and 1
        ##################################

        if not 0 <= self.spin_reserve_factor <= 1:
            raise ValueError("PowNet: Spin reserve factor must be between 0 and 1.")
        if not 0 <= self.line_loss_factor <= 1:
            raise ValueError("PowNet: Line loss factor must be between 0 and 1.")
        if not 0 <= self.line_capacity_factor <= 1:
            raise ValueError("PowNet: Line capacity factor must be between 0 and 1.")

        ##################################
        # Timeseries have the correct length
        ##################################

        if len(self.demand) != self.num_sim_hours:
            raise ValueError(
                f"PowNet: Demand timeseries must be of length {self.num_sim_hours} but got {len(self.demand)}."
            )

        attrs_to_check = [
            "solar_capacity",
            "wind_capacity",
            "import_capacity",
            "hydro_capacity",
            "susceptance",
            "line_capacity",
        ]
        for attr in attrs_to_check:
            temp_df = getattr(self, attr)
            if (not temp_df.empty) and (len(temp_df) != self.num_sim_hours):
                raise ValueError(
                    f"PowNet: {attr} must be of length {self.num_sim_hours} but got {len(temp_df)}."
                )

        if len(self.daily_hydro_capacity) not in [0, self.num_sim_days]:
            raise ValueError(
                f"PowNet: Daily hydropower timeseries must be of length {self.num_sim_days}."
            )

        ##################################
        # The derated capacities of thermal units must be above its minimum capacity
        ##################################

        if not (self.thermal_derated_capacity >= self.thermal_min_capacity).all().all():
            # Identify units with derated capacity below the minimum capacity
            units_below_min_capacity = (
                self.thermal_derated_capacity[
                    self.thermal_derated_capacity < self.thermal_min_capacity
                ]
                .stack()
                .index.tolist()
            )
            raise ValueError(
                f"PowNet: The derated capacity of thermal units must be above the minimum capacity:\n{units_below_min_capacity}"
            )

        ##################################
        # Spinning reserve cannot be larger than the whole system's demand
        ##################################

        if self.spin_reserve_mw is not None:
            if (self.spin_reserve_mw > self.demand.sum(axis=1)).any():
                raise ValueError(
                    "PowNet: Spin reserve cannot be larger than demand at any time."
                )

        ##################################
        # Consistency in the number of units
        ##################################

        # The number of columns contracts must equal the number of generators
        number_of_non_fossil_generators = len(
            self.hydro_unit_node
            | self.daily_hydro_unit_node
            | self.solar_unit_node
            | self.wind_unit_node
            | self.import_unit_node
        )

        # Number of nondispatch contracts
        if len(self.nondispatch_contracts) != number_of_non_fossil_generators:
            raise ValueError(
                f"PowNet: The number of non-dispatchable contracts must equal the number of non-fossil generators. {len(self.nondispatch_contracts)} != {number_of_non_fossil_generators}"
            )

        number_of_generators = (
            len(self.thermal_unit_node) + number_of_non_fossil_generators
        )
        if len(self.fuelmap) != number_of_generators:
            raise ValueError(
                f"PowNet: The number of units in the fuelmap must be equal to the number of generators. {len(self.fuelmap)} != {number_of_generators}"
            )

        # Number of nodes in the node_generator dictionary must be equal to the number of nodes
        if len(self.node_generator) != len(self.nodes):
            raise ValueError(
                f"PowNet: The number of nodes in the node_generator dictionary must be equal to the number of nodes. {len(self.node_generator)} != {len(self.nodes)}"
            )

        # Number of generators in the node_generator dictionary must be equal to the number of generators
        if sum(len(v) for v in self.node_generator.values()) != number_of_generators:
            raise ValueError(
                f"PowNet: The number of generators in the node_generator dictionary must be equal to the number of generators. {sum(len(v) for v in self.node_generator.values())} != {number_of_generators}"
            )

        # Number of nodes in the node_edge dictionary must be equal to the number of nodes
        if len(self.node_edge) != len(self.nodes):
            raise ValueError(
                f"PowNet: The number of nodes in the node_edge dictionary must be equal to the number of nodes. {len(self.node_edge)} != {len(self.nodes)}"
            )

        # Number of edges in the node_edge dictionary must be equal to TWICE the number of edges
        if sum(len(v) for v in self.node_edge.values()) != 2 * len(self.edges):
            raise ValueError(
                f"PowNet: The number of edges in the node_edge dictionary must be equal to the number of edges. {sum(len(v) for v in self.node_edge.values())} != {2*len(self.edges)}"
            )

        ##################################
        # Names cannot repeat across different types
        ##################################
        if self.all_generators.intersection(self.demand_nodes):
            raise ValueError(
                "PowNet: Generator names cannot be the same as the name of demand nodes."
            )
        if set(self.ess_unit_attach.keys()).intersection(self.all_generators):
            raise ValueError(
                "PowNet: Energy storage names cannot be the same as the name of generators."
            )

        ##################################
        # ESS must be connected to either a node or a generator
        ##################################

        assigned_ess = (
            list(self.ess_hydro_units.keys())
            + list(self.ess_daily_hydro_units.keys())
            + list(self.ess_solar_units.keys())
            + list(self.ess_wind_units.keys())
            + list(self.ess_thermal_units.keys())
            + list(self.ess_substation_units.keys())
        )
        if len(assigned_ess) != len(self.ess_unit_attach):
            raise ValueError(
                "PowNet: Energy storage systems must be connected to either a node or a generator."
            )

    def print_summary(self):
        input_summary = textwrap.dedent(
            f"""
        \n\nPowNet Input Data Summary:
        {'Timestamp':<25} = {self.timestamp}
        {'Model name':<25} = {self.model_name}
        {'Year':<25} = {self.year}
        ---- System characteristics ----
        {'No. of nodes':<25} = {len(self.nodes)}
        {'No. of edges':<25} = {len(self.edges)}
        {'No. of thermal units':<25} = {len(self.thermal_unit_node)}
        {'No. of demand nodes':<25} = {len(self.demand_nodes)}
        {'Peak demand':<25} = {round(self.demand.sum(axis=1).max())} MW

        ---- Renewable capacities ----
        {'Hydropower units':<25} = {len(self.hydro_unit_node)}
        {'Daily hydropower units':<25} = {len(self.daily_hydro_unit_node)}
        {'Solar units':<25} = {len(self.solar_unit_node)}
        {'Wind units':<25} = {len(self.wind_unit_node)}
        {'Import units':<25} = {len(self.import_unit_node)}

        ---- Energy storage ----
        {'No. of hydropower with ESS':<25} = {len(self.ess_hydro_units)}
        {'No. of daily hydropower with ESS':<25} = {len(self.ess_daily_hydro_units)}
        {'No. of Solar with ESS':<25} = {len(self.ess_solar_units)}
        {'No. of Wind with ESS':<25} = {len(self.ess_wind_units)}
        {'No. of Thermal units with ESS':<25} = {len(self.ess_thermal_units)}
        {'No. of Grid ESS':<25} = {len(self.ess_substation_units)}

        ---- Modeling parameters ----
        {'Simulation horizon':<25} = {self.sim_horizon} hours
        {'Number of simulation days':<25} = {self.num_sim_days}
        {'Use spin variable':<25} = {self.use_spin_var}
        {'Power flow':<25} = {self.dc_opf}
        {'Spin reserve factor:':<25} = {self.spin_reserve_factor if self.spin_reserve_mw is None else 'Use an absolute value in MW.'}
        {'Spin reserve amount (MW):':<25} = {self.spin_reserve_mw if self.spin_reserve_mw is not None else 'Using a factor.'}
        {'Line loss factor':<25} = {self.line_loss_factor}
        {'Line capacity factor':<25} = {self.line_capacity_factor}
        {'Load shortfall penalty':<25} = {self.load_shortfall_penalty_factor}
        {'Reserve shortfall penalty':<25} = {self.spin_shortfall_penalty_factor}\n

        """
        )
        logger.warning(input_summary)

    def load_and_check_data(self):
        """Load and check the input data."""
        self.load_data()
        self.check_data()
        self.print_summary()

    def update_hydro_capacity(self, hydropower: pd.DataFrame):
        """Update the hydropower timeseries."""
        # Check that the dimension of the hydropower timeseries remains the same
        if len(hydropower) != len(self.hydro_capacity):
            raise ValueError(
                "PowNet: The length of the hydropower timeseries must remain the same."
            )
        # Check that all hydropower units are present
        if set(hydropower.columns) != set(self.hydro_capacity.columns):
            raise ValueError(
                "PowNet: The hydropower timeseries must contain all hydropower units."
            )
        self.hydro_capacity = hydropower.copy()

    def get_unit_contracts(self) -> dict[str, str]:
        temp_dict = self.fuel_contracts.copy()
        temp_dict.update(self.nondispatch_contracts)
        temp_dict.update(self.ess_contracts)
        return temp_dict



import sys
import os

sys.path.insert(0, os.path.abspath("../../src/"))

# Import mock modules for readthedocs
autodoc_mock_imports = [
    "datetime",
    "pandas",
    "gurobipy",
    "highspy",
    "networkx",
    "numpy",
    "matplotlib.pyplot",
    "math",
    "__future__",
    "re",
]

# -- Project information

project = "PowNet"
copyright = "2024, Critical Infrastructure Systems (CIS) Lab, Cornell University"
author = "Critical Infrastructure Systems Lab (CIS), Cornell University"

release = "2.0"
version = "2.0.0"

# -- General configuration

github_username = "HishamEldardiry"
github_repository = "PowNet_Documentation"


extensions = [
    "sphinx.ext.duration",
    "sphinx.ext.doctest",
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx_autodoc_typehints",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
    "sphinx.ext.mathjax",
    "sphinx.ext.viewcode",
    "sphinx_rtd_theme",
    "sphinx.ext.viewcode",
    "sphinx_toolbox",
]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "sphinx": ("https://www.sphinx-doc.org/en/master/", None),
}
intersphinx_disabled_domains = ["std"]

templates_path = ["_templates"]
html_static_path = ["_static"]

# -- Options for HTML output

html_theme = "sphinx_rtd_theme"
html_theme_options = {
    "version": "",
    "release": "",
}

html_context = {
    "display_github": True,  # Add 'Edit on Github' link instead of 'View page source'
    "last_updated": False,
    "commit": False,
}

# -- html_show_sphinx = False

# -- Options for EPUB output
epub_show_urls = "footnote"


# -- customize the CSS styling
def setup(app):
    app.add_css_file("custom.css")



"""
model.py: PowerSystemModel is wrapper of gurobipy.Model and highspy.Highs
to provide a more user-friendly interface while compatible with HiGHs solver.
"""

import os

import gurobipy as gp
import highspy
import pandas as pd

from pownet.data_utils import (
    get_node_hour_from_flow_constraint,
    parse_lmp,
    parse_node_variables,
)

from .rounding_algo import optimize_with_rounding


class PowerSystemModel:
    def __init__(self, model: gp.Model):
        self.model = model
        self.solver: str = "gurobi"

        # Rounding related variables
        self.status_vars: gp.tupledict = None

        # Define dictionaries of functions for Gurobi and HiGHs
        self.optimize_functions = {
            "gurobi": self._optimize_gurobi,
            "highs": self._optimize_highs,
        }
        self.check_feasible_functions = {
            "gurobi": self._check_feasible_gurobi,
            "highs": self._check_feasible_highs,
        }
        self.get_objval_functions = {
            "gurobi": self._get_objval_gurobi,
            "highs": self._get_objval_highs,
        }
        self.get_status_functions = {
            "gurobi": self._get_status_gurobi,
            "highs": self._get_status_highs,
        }
        self.get_solution_functions = {
            "gurobi": self.get_solution_gurobi,
            "highs": self.get_solution_highs,
        }
        self.get_runtime_functions = {
            "gurobi": self.get_runtime_gurobi,
            "highs": self.get_runtime_highs,
        }
        self.solve_for_lmp_functions = {
            "gurobi": self.solve_for_lmp_gurobi,
            "highs": self.solve_for_lmp_highs,
        }

        self.rounding_optimization_time: float = None
        self.rounding_iterations: int = None

    def write_mps(self, output_folder: str, filename: str):
        if not isinstance(self.model, gp.Model):
            raise ValueError("The model must be a Gurobi model")
        if not os.path.exists(output_folder):
            os.makedirs(output_folder)
        self.model.write(os.path.join(output_folder, f"{filename}.mps"))

    def _optimize_gurobi(
        self,
        log_to_console: bool,
        mipgap: float,
        timelimit: int,
        num_threads: int,
    ):

        self.model.Params.LogToConsole = log_to_console
        self.model.Params.MIPGap = mipgap
        self.model.Params.TimeLimit = timelimit
        self.model.Params.Threads = num_threads

        self.model.optimize()

    def _optimize_highs(
        self, log_to_console: bool, mipgap: float, timelimit: int, num_threads: int
    ):
        # Export the instance to MPS and solve with HiGHs
        mps_file = "temp_instance_for_HiGHs.mps"
        self.model.write(mps_file)
        self.model = highspy.Highs()
        self.model.readModel(mps_file)

        self.model.setOptionValue("log_to_console", log_to_console)
        self.model.setOptionValue("mip_rel_gap", mipgap)
        self.model.setOptionValue("time_limit", timelimit)
        self.model.setOptionValue("threads", num_threads)

        # Must faster than automatic selection
        self.model.setOptionValue("solver", "simplex")

        self.model.run()
        # Delete the MPS file
        os.remove(mps_file)

    def optimize(
        self,
        solver: str = "gurobi",
        log_to_console: bool = True,
        mipgap: float = 1e-3,
        timelimit: int = 600,
        num_threads: int = 0,
    ):
        if solver not in ["gurobi", "highs"]:
            raise ValueError("The solver must be either 'gurobi' or 'highs'")

        # Update the solver attribute for referencing in other methods
        self.solver = solver
        self.optimize_functions[self.solver](
            log_to_console=log_to_console,
            mipgap=mipgap,
            timelimit=timelimit,
            num_threads=num_threads,
        )

    def optimize_with_rounding(
        self,
        rounding_strategy: str,
        max_rounding_iter: int,
        threshold: float = 0,
        mipgap: float = 1e-3,
        timelimit: int = 600,
        num_threads: int = 0,
        log_to_console: bool = False,
    ) -> None:
        self.model, self.rounding_optimization_time, self.rounding_iterations = (
            optimize_with_rounding(
                model=self.model,
                rounding_strategy=rounding_strategy,
                threshold=threshold,
                max_rounding_iter=max_rounding_iter,
                log_to_console=log_to_console,
                mipgap=mipgap,
                timelimit=timelimit,
                num_threads=num_threads,
            )
        )

    def _check_feasible_gurobi(self) -> bool:
        not_allowed_statuses = [
            gp.GRB.Status.INFEASIBLE,
            gp.GRB.Status.INF_OR_UNBD,
            gp.GRB.Status.UNBOUNDED,
        ]
        return self.model.status not in not_allowed_statuses

    def _check_feasible_highs(self) -> bool:
        model_status = self.model.getModelStatus()
        return self.model.modelStatusToString(model_status) == "Optimal"

    def check_feasible(self) -> bool:
        return self.check_feasible_functions[self.solver]()

    def write_ilp_mps(self, output_folder: str, instance_name: str):
        # Write the infeasible model to a file
        self.model.computeIIS()
        ilp_file = os.path.join(
            output_folder,
            f"infeasible_{instance_name}.ilp",
        )
        self.model.write(ilp_file)
        # Write the infeasible model to a MPS file
        self.model.write(ilp_file.replace(".ilp", ".mps"))

    def _get_objval_gurobi(self) -> float:
        return self.model.objVal

    def _get_objval_highs(self) -> float:
        info = self.model.getInfo()
        return info.objective_function_value

    def get_objval(self) -> float:
        return self.get_objval_functions[self.solver]()

    def _get_status_gurobi(self):
        return self.model.status

    def _get_status_highs(self):
        return self.model.getModelStatus()

    def get_status(self):
        return self.get_status_functions[self.solver]()

    def get_model(self):
        return self.model

    def get_solution_gurobi(self) -> dict:
        return {
            "varname": self.model.getAttr("varname"),
            "value": self.model.getAttr("X"),
        }

    def get_solution_highs(self) -> dict:
        return {
            "varname": [
                self.model.getColName(i)[1]
                for i in range(self.model.getNumCol())  # getColName returns a tuple
            ],
            "value": self.model.getSolution().col_value,
        }

    def get_solution(self) -> pd.DataFrame:
        return pd.DataFrame(self.get_solution_functions[self.solver]())

    def get_runtime_gurobi(self) -> float:
        return self.model.Runtime

    def get_runtime_highs(self) -> float:
        return self.model.getRunTime()

    def get_runtime(self) -> float:
        return self.get_runtime_functions[self.solver]()

    def solve_for_lmp_gurobi(self) -> dict:
        """Return the locational marginal price (LMP).

        Args:
            model: The Gurobi model.

        Returns:
            The LMP at each node.
        """
        # Fix the binary variables
        model_fixed = self.model.fixed()
        # Reoptimize as linear program, which is already obtained by fixing the binary variables
        model_fixed.optimize()
        # Get the dual variables
        pi = model_fixed.getAttr("Pi", model_fixed.getConstrs())
        # Filter to only constraints that are nodal balance, which has 'flowBal' in the name
        nodal_price = {
            constr.ConstrName: pi[i]
            for i, constr in enumerate(model_fixed.getConstrs())
            if "flowBal" in constr.ConstrName
        }
        return nodal_price

    def solve_for_lmp_highs(self) -> dict:
        raise NotImplementedError("This method is not implemented for HiGHs solver")

    def solve_for_lmp(self) -> dict:
        return self.solve_for_lmp_functions[self.solver]()

    def solve_for_export_capacity(
        self, shared_nodes: list, sim_horizon: int, step_k: int
    ) -> tuple:
        """Return the export capacity and hourly prices at the shared nodes"""
        # Fix binary variables to simulate fixing unit commitments
        model_fixed = self.model.fixed()

        # Add export variables to the model with a negative coefficient (minimization problem)
        # The value should be high enough to incentivize export but not high enough to create shortfall
        export_vars = model_fixed.addVars(
            shared_nodes,
            range(1, sim_horizon + 1),
            vtype=gp.GRB.CONTINUOUS,
            obj=-1,  # A small negative value should urge the model to export
            name="export",
        )

        # Add export variables to the flow balance at the shared nodes
        for constr in model_fixed.getConstrs():
            node, t = get_node_hour_from_flow_constraint(constr.ConstrName)
            if (node is not None) and (node in shared_nodes):
                model_fixed.chgCoeff(
                    constr,
                    export_vars[(node, t)],
                    -1,  # This will add the export variable to the demand on the RHS
                )
        model_fixed.optimize()

        # Extract the export capacity
        export_capacity = pd.DataFrame(
            {
                "varname": [v.varname for v in export_vars.values()],
                "value": [v.x for v in export_vars.values()],
            }
        )

        # Format the dataframe with additional columns
        export_capacity = parse_node_variables(export_capacity, sim_horizon, step_k)

        return export_capacity.pivot(index="hour", columns="node", values="value")

    def solve_for_export_prices(
        self, shared_nodes: list, sim_horizon: int, step_k: int
    ) -> pd.DataFrame:
        """The export prices are locational marginal prices at the shared nodes."""
        export_prices = parse_lmp(
            lmp=self.solve_for_lmp(),
            sim_horizon=sim_horizon,
            step_k=step_k,
        )
        export_prices = export_prices[export_prices["node"].isin(shared_nodes)]
        return export_prices.pivot(index="hour", columns="node", values="value")



""" This is test_data_processor.py
"""

import os
import unittest
from pownet.core.data_processor import (
    DataProcessor,
)

from pownet.folder_utils import get_model_dir


class TestDataProcessor(unittest.TestCase):
    def test_initialization(self):
        # Arrange
        test_model_library_path = os.path.abspath(
            os.path.join(os.path.dirname(__file__), "..", "test_model_library")
        )
        model_name = "dummy"
        year = 2024
        frequency = 50

        # Act
        processor = DataProcessor(
            input_folder=test_model_library_path,
            model_name=model_name,
            year=year,
            frequency=frequency,
        )

        processor.execute_data_pipeline()

        # Assert
        self.assertEqual(processor.model_name, model_name)
        self.assertEqual(processor.year, year)
        self.assertEqual(processor.frequency, frequency)
        self.assertEqual(processor.wavelength, 6000)
        self.assertEqual(
            processor.model_folder,
            os.path.join(get_model_dir(), test_model_library_path, model_name),
        )
        # Timeseries should have 8760 rows
        self.assertEqual(processor.cycle_map, {})
        self.assertEqual(processor.thermal_derate_factors.shape[0], 8760)


if __name__ == "__main__":
    unittest.main()



""" This is the reservoir module.
"""

from .reservoir import Reservoir
from .basin import Basin
from .reservoir_operator import ReservoirOperator
from .solve_release import (
    solve_release_from_target_storage,
    solve_release_from_dispatch,
)



"""
This script uses the Gurobi solver to solve for release.
"""

import gurobipy as gp
import pandas as pd


def solve_release_from_target_storage(
    reservoir_name: str,
    start_day: int,
    end_day: int,
    max_release: float,
    max_storage: float,
    initial_storage: float,
    target_storage: pd.Series,
    min_flow: pd.Series,
    total_inflow: pd.Series,
) -> tuple[pd.Series, pd.Series, pd.Series, float]:
    """Build an optimization problem to find the optimal release from the reservoir.
    The objective is to minimize the storage deviation from the target storage with L1 norm.

    OBJECTIVE FUNCTION:
    min | target_storage - storage |

    Rewrite this as:

    min sum_{day} sbar[day]
    s.t.
    - sbar[day] <= target_storage[day] - storage[day] <= sbar[day]

    CONSTRAINTS:

    """
    model = gp.Model(f"match_storage_{reservoir_name}")
    model.setParam("OutputFlag", 0)

    timesteps = range(start_day, end_day + 1)

    # Create the decision variables
    release_vars = model.addVars(
        timesteps,
        lb=0,
        ub=max_release,
        name="release",
    )
    spill_vars = model.addVars(
        timesteps,
        lb=0,
        name="spill",
    )
    storage_vars = model.addVars(
        timesteps,
        lb=0,
        ub=max_storage,
        name="storage",
    )
    # The deviation from target storage is represented by sbar
    sbar = model.addVars(
        timesteps,
        lb=0,
        name="sbar",
    )

    """
    When using max/min as a function, we need to use gp.max_ and gp.min_
    see https://support.gurobi.com/hc/en-us/community/posts/360078185112-gurobipy-Model-addGenConstrMin-Invalid-data-in-vars-array

    spills[day] = max(0, total_inflow[day] + storage[day] - max_storage - release[day])
    spills[day] = gp.max_(0, spill_bar[day])

    with
    spill_bar[day] = total_inflow[day] + storage[day] - max_storage - release[day]
    """
    spill_bar = model.addVars(
        timesteps,
        lb=-gp.GRB.INFINITY,
        name="spill_bar",
    )

    # Create the objective function
    model.setObjective(
        gp.quicksum(sbar[day] for day in timesteps),
        sense=gp.GRB.MINIMIZE,
    )

    # Lower and upper bounds for the storage deviation are defined by sbar
    model.addConstrs(
        (
            -1 * sbar[day] <= target_storage[day] - storage_vars[day]
            for day in timesteps
        ),
        name="c_min_sbar",
    )
    model.addConstrs(
        (target_storage[day] - storage_vars[day] <= sbar[day] for day in timesteps),
        name="c_max_sbar",
    )
    # Minimum release has not been enforced when defining
    # the variable.
    model.addConstrs(
        (release_vars[day] >= min_flow[day] for day in timesteps),
        name="c_min_release",
    )
    # Define spill
    model.addConstrs(
        (spill_vars[day] == gp.max_(0, spill_bar[day]) for day in timesteps),
        name="c_spill",
    )
    # Define spill_bar
    model.addConstrs(
        (
            spill_bar[day]
            == total_inflow[day] + storage_vars[day] - max_storage - release_vars[day]
            for day in timesteps
        ),
        name="c_define_spill_bar",
    )

    # The storage at the start day is the initial storage
    model.addConstr(
        (
            storage_vars[start_day]
            == initial_storage
            + total_inflow[start_day]
            - release_vars[start_day]
            - spill_vars[start_day]
        ),
        name="c_initial_storage",
    )

    # Storage at other days
    model.addConstrs(
        (
            storage_vars[day]
            == storage_vars[day - 1]
            + total_inflow[day]
            - release_vars[day]
            - spill_vars[day]
            for day in range(start_day + 1, end_day + 1)
        ),
        name="c_storage",
    )

    # Solve the optimization problem
    model.optimize()

    # Check the status of the optimization
    if model.status == gp.GRB.INFEASIBLE:
        # Export IIS file if the problem is infeasible
        model.computeIIS()
        model.write(f"release_from_storage_{reservoir_name}.ilp")
        raise Exception(f"Match storage is infeasible: {reservoir_name}")

    # Get the solution
    opt_release = pd.Series(
        [release_vars[i].x for i in release_vars],
        name="release",
        index=timesteps,
    )
    opt_spill = pd.Series(
        [spill_vars[i].X for i in spill_vars],
        name="spill",
        index=timesteps,
    )
    opt_storage = pd.Series(
        [storage_vars[i].X for i in storage_vars],
        name="storage",
        index=timesteps,
    )
    return opt_release, opt_spill, opt_storage, model.objVal


def solve_release_from_dispatch(
    reservoir_name: str,
    daily_dispatch: float,
    turbine_factor: float,
    max_head: float,
    max_level: float,
    min_level: float,
    level_t0: float,
    storage_max: float,
    storage_t0: float,
    inflow: float,
    min_release: float,
    max_release: float,
    max_generation: float,
) -> tuple[float, float, float, float]:
    """
    For each day, solve for release_t from daily dispatch_t as an optimization problem.

    min         abs(DISPATCH_t - daily_hydropower_t)
    s.t.

    1. Calculate hourly_hydropower_t
        hourly_hydropower_t = TURBINE_FACTOR * DENSITY * g * head_t * release_t

            where
                    DENSITY = 998  # kg/m3
                    g = 9.81  # m/s2

    2. Hydrautic head is a function of water levvel
    head_t = MAX_HEAD - (MAX_LEVEL - mid_level_t)

    where mid_level_t = (level_t0 + level_t) / 2

    3. level_t is a function of storage
    level_t = storage_t / STORAGE_MAX * (MAX_LEVEL - MIN_LEVEL) + MIN_LEVEL

    4. Definition of storage
    storage_t = STORAGE_t0 + INFLOW_t - release_t - spill_t

    5. Definition of spill
    spill_t = max(0, INFLOW_t + storage_t - STORAGE_MAX - release_t)
    Rearrange to
    spill_t = max(0, spill_bar)

    6. Definition of spill_bar
    spill_bar = INFLOW_t + storage_t - STORAGE_MAX - release_t

    Note that the objective function is not linear because of the absolute value.
    To make it linear, we introduce a new variable mismatch_t and rewrite
    the objective function as:

    min     mismatch

    s.t.

            -mismatch <= DISPATCH_t - daily_hydropower_t <= mismatch

    """
    ############################
    # Creating the model
    ############################
    model = gp.Model(f"get_release_from_dispatch_{reservoir_name}")
    model.setParam("OutputFlag", 0)

    # Create variables
    mismatch = model.addVar(lb=0.0, name="mismatch")
    release = model.addVar(lb=min_release, ub=max_release, name="release")

    spill = model.addVar(lb=0.0, name="spill")
    spill_bar = model.addVar(lb=-gp.GRB.INFINITY, name="spill_bar")

    level = model.addVar(lb=min_level, ub=max_level, name="level")
    storage = model.addVar(lb=0.0, ub=storage_max, name="storage")

    head = model.addVar(lb=0.0, ub=max_head, name="head")

    # Unbounded hydropower
    unb_hourly_hydropower = model.addVar(lb=0.0, name="unb_hourly_hydropower")
    # Bounded hydropower
    hourly_hydropower = model.addVar(lb=0.0, name="hourly_hydropower")
    daily_hydropower = model.addVar(lb=0.0, name="daily_hydropower")

    # Set objective
    model.setObjective(mismatch, gp.GRB.MINIMIZE)

    # Bounds of mismatch
    model.addConstr(
        -mismatch <= daily_dispatch - daily_hydropower, name="c_mismatch_lb"
    )
    model.addConstr(mismatch >= daily_dispatch - daily_hydropower, name="c_mismatch_ub")

    # (1) Define hydropower. Note that the equation requires
    # converting the release from m3/day to m3/s.
    density = 998  # kg/m3
    gravity = 9.81  # m/s2
    num_seconds_in_day = 24 * 3600

    model.addConstr(
        unb_hourly_hydropower
        == turbine_factor
        * density
        * gravity
        * head
        * release
        / num_seconds_in_day
        / 1e6,  # Convert to MW
        name="c_unbounded_hydropower",
    )

    # (1a) Limit hourly hydropower by turbine capacity
    model.addConstr(
        hourly_hydropower == gp.min_(unb_hourly_hydropower, constant=max_generation),
        name="c_hydropower",
    )

    # (1b) Daily_hydropower in MW-day
    model.addConstr(
        daily_hydropower == hourly_hydropower * 24,
        name="c_daily_hydropower",
    )

    # (2) Define hydraulic head as a function of mid_level
    model.addConstr(
        head == max_head - (max_level - (level_t0 + level) / 2),
        name="c_head",
    )

    # (3) Define level
    model.addConstr(
        level == storage / storage_max * (max_level - min_level) + min_level,
        name="c_level",
    )

    # (4) Define storage
    model.addConstr(
        storage == storage_t0 + inflow - release - spill,
        name="c_storage",
    )

    # (5) Define spill_bar
    model.addConstr(
        spill_bar == inflow + storage - storage_max - release,
        name="c_spill_bar",
    )

    # (6) Define spill
    model.addConstr(spill == gp.max_(0, spill_bar), name="c_spill")

    ############################
    # Solve the optimization problem
    ############################
    model.optimize()

    # Check the status of the optimization
    if model.status == gp.GRB.INFEASIBLE:
        # Export IIS file if the problem is infeasible
        model.computeIIS()
        model.write("match_dispatch.ilp")
        print(f"\nMatching dispatch is infeasible: {reservoir_name}")

    # Get the solution
    release_t = release.X
    spill_t = spill.X
    storage_t = storage.X
    level_t = level.X
    hourly_hydropower_t = hourly_hydropower.X
    daily_hydropower_t = daily_hydropower.X
    mismatch_t = mismatch.X

    return (
        release_t,
        spill_t,
        storage_t,
        level_t,
        hourly_hydropower_t,
        daily_hydropower_t,
        mismatch_t,
    )
